#! @PYSHEBANG@
# -*- coding: utf-8 -*-

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

from __future__ import print_function, division

import sys
import os
import getopt
import time
import socket

try:
    import ntp.packet
    import ntp.util
    import ntp.agentx_packet
    ax = ntp.agentx_packet
    from ntp.agentx import PacketControl
except ImportError as e:
    sys.stderr.write(
        "ntpsnmpd: can't find Python NTP library.\n")
    sys.stderr.write("%s\n" % e)
    sys.exit(1)


# TODO This is either necessary, or a different workaround is.
ntp.util.deunicode_units()

logfp = sys.stderr
nofork = False
debug = 0
defaultTimeout = 30

log = (lambda msg, msgdbg: ntp.util.dolog(logfp, msg, debug, msgdbg))

ntpRootOID = (1, 3, 6, 1, 2, 1, 197)  # mib-2 . 197, aka: NTPv4-MIB

snmpTrapOID = (1, 3, 6, 1, 6, 3, 1, 1, 4, 1, 0)
snmpSysUptime = (1, 3, 6, 1, 2, 1, 1, 3, 0)

DEFHOST = "localhost"
DEFLOG = "ntpsnmpd.log"


class DataSource(ntp.agentx.MIBControl):
    def __init__(self, hostname=DEFHOST, settingsFile=None, notifySpin=0.1):
        # This is defined as a dict tree because it is simpler, and avoids
        # certain edge cases
        # OIDs are relative from ntp root
        ntp.agentx.MIBControl.__init__(self, mibRoot=ntpRootOID)
        # MIB node init
        # block 0
        self.addNode((0,))  # ntpEntNotifications
        self.addNode((0, 1))  # ntpEntNotifModeChange
        self.addNode((0, 2))  # ntpEntNotifStratumChange
        self.addNode((0, 3))  # ntpEntNotifSyspeerChange
        self.addNode((0, 4))  # ntpEntNotifAddAssociation
        self.addNode((0, 5))  # ntpEntNotifRemoveAsociation
        self.addNode((0, 6))  # ntpEntNotifConfigChanged
        self.addNode((0, 7))  # ntpEntNotifLeapSecondAnnounced
        self.addNode((0, 8))  # ntpEntNotifHeartbeat
        # block 1
        # block 1, 1
        self.addNode((1, 1, 1, 0),  # ntpNetSoftwareName utf8str
                     (lambda oid: self.cbr_systemInfo(oid, "name")))
        self.addNode((1, 1, 2, 0),  # ntpEntSoftwareVersion utf8str
                     (lambda oid: self.cbr_systemInfo(oid, "version")))
        self.addNode((1, 1, 3, 0),  # ntpEntSoftwareVendor utf8str
                     (lambda oid: self.cbr_systemInfo(oid, "vendor")))
        self.addNode((1, 1, 4, 0),  # ntpEntSystemType utf8str
                     (lambda oid: self.cbr_systemInfo(oid, "system")))
        self.addNode((1, 1, 5, 0),  # ntpEntTimeResolution uint32
                     self.cbr_timeResolution)
        self.addNode((1, 1, 6, 0),  # ntpEntTimePrecision int32
                     self.cbr_timePrecision)
        self.addNode((1, 1, 7, 0),  # ntpEntTimeDistance DisplayString
                     self.cbr_timeDistance)
        # block 1, 2
        self.addNode((1, 2, 1, 0),  # ntpEntStatusCurrentMode INTEGER {...}
                     self.cbr_statusCurrentMode)
        self.addNode((1, 2, 2, 0),  # ntpEntStatusStratum NtpStratum
                     self.cbr_statusStratum)
        self.addNode((1, 2, 3, 0),  # ntpEntStatusActiveRefSourceId uint32
                     self.cbr_statusActiveRefSourceID)
        self.addNode((1, 2, 4, 0),  # ntpEntStatusActiveRefSourceName utf8str
                     self.cbr_statusActiveRefSourceName)
        self.addNode((1, 2, 5, 0),  # ntpEntStatusActiveOffset DisplayString
                     self.cbr_statusActiveOffset)
        self.addNode((1, 2, 6, 0),  # ntpEntStatusNumberOfRefSources unit32
                     self.cbr_statusNumRefSources)
        self.addNode((1, 2, 7, 0),  # ntpEntStatusDispersion DisplayString
                     self.cbr_statusDispersion)
        self.addNode((1, 2, 8, 0),  # ntpEntStatusEntityUptime TimeTicks
                     self.cbr_statusEntityUptime)
        self.addNode((1, 2, 9, 0),  # ntpEntStatusDateTime NtpDateTime
                     self.cbr_statusDateTime)
        self.addNode((1, 2, 10, 0),  # ntpEntStatusLeapSecond NtpDateTime
                     self.cbr_statusLeapSecond)
        self.addNode((1, 2, 11, 0),  # ntpEntStatusLeapSecondDirection int32
                     self.cbr_statusLeapSecDirection)
        self.addNode((1, 2, 12, 0),  # ntpEntStatusInPkts Counter32
                     self.cbr_statusInPkts)
        self.addNode((1, 2, 13, 0),  # ntpEntStatusOutPkts Counter32
                     self.cbr_statusOutPkts)
        self.addNode((1, 2, 14, 0),  # ntpEntStatusBadVersion Counter32
                     self.cbr_statusBadVersion)
        self.addNode((1, 2, 15, 0),  # ntpEntStatusProtocolError Counter32
                     self.cbr_statusProtocolError)
        self.addNode((1, 2, 16, 0),  # ntpEntStatusNotifications Counter32
                     self.cbr_statusNotifications)
        self.addNode((1, 2, 17, 1, 1))  # ntpEntStatPktMode INTEGER {...}
        self.addNode((1, 2, 17, 1, 2))  # ntpEntStatPktSent Counter32
        self.addNode((1, 2, 17, 1, 3))  # ntpEntStatPktReceived Counter32
        # block 1, 3
        self.addNode((1, 3, 1, 1, 1),  # ntpAssocId uint32 (1..99999)
                     dynamic=self.sub_assocID)
        self.addNode((1, 3, 1, 1, 2),  # ntpAssocName utf8str
                     dynamic=self.sub_assocName)
        self.addNode((1, 3, 1, 1, 3),  # ntpAssocRefId DisplayString
                     dynamic=self.sub_assocRefID)
        self.addNode((1, 3, 1, 1, 4),  # ntpAssocAddressType InetAddressType
                     dynamic=self.sub_assocAddrType)
        self.addNode((1, 3, 1, 1, 5),  # ntpAssocAddress InetAddress SIZE
                     dynamic=self.sub_assocAddr)
        self.addNode((1, 3, 1, 1, 6),  # ntpAssocOffset DisplayString
                     dynamic=self.sub_assocOffset)
        self.addNode((1, 3, 1, 1, 7),  # ntpAssocStratum NtpStratum
                     dynamic=self.sub_assocStratum)
        self.addNode((1, 3, 1, 1, 8),  # ntpAssocStatusJitter DisplayString
                     dynamic=self.sub_assocJitter)
        self.addNode((1, 3, 1, 1, 9),  # ntpAssocStatusDelay DisplayString
                     dynamic=self.sub_assocDelay)
        self.addNode((1, 3, 1, 1, 10),  # ntpAssocStatusDispersion DisplayStr
                     dynamic=self.sub_assocDispersion)
        self.addNode((1, 3, 2, 1, 1),  # ntpAssocStatInPkts Counter32
                     dynamic=self.sub_assocStatInPkts)
        self.addNode((1, 3, 2, 1, 2),  # ntpAssocStatOutPkts Counter32
                     dynamic=self.sub_assocStatOutPkts)
        self.addNode((1, 3, 2, 1, 3),  # ntpAssocStatProtocolError Counter32
                     dynamic=self.sub_assocStatProtoErr)
        # block 1, 4
        self.addNode((1, 4, 1, 0),  # ntpEntHeartbeatInterval unit32
                     self.cbr_entHeartbeatInterval,
                     self.cbw_entHeartbeatInterval)
        self.addNode((1, 4, 2, 0),  # ntpEntNotifBits BITS {...}
                     self.cbr_entNotifBits,
                     self.cbw_entNotifBits)
        # block 1, 5
        self.addNode((1, 5, 1, 0),  # ntpEntNotifMessage utf8str
                     self.cbr_entNotifMessage)
        # block 2 # all compliance statements
        # print(repr(self.oidTree))
        # print(self.oidTree[1]["subids"][1][1][0])
        self.session = ntp.packet.ControlSession()
        self.hostname = hostname if hostname else DEFHOST
        self.session.openhost(self.hostname)
        self.settingsFilename = settingsFile
        # Cache so we don't hammer ntpd, default 1 second timeout
        # Timeout default pulled from a hat: we don't want it to last for
        # long, just not flood ntpd with duplicatte requests during a walk.
        self.cache = ntp.util.Cache(1)
        self.oldValues = {}  # Used by notifications to detect changes
        # spinGap so we don't spam ntpd with requests during notify checks
        self.notifySpinTime = notifySpin
        self.lastNotifyCheck = 0
        self.lastHeartbeat = 0  # Timestamp used for heartbeat notifications
        self.heartbeatInterval = 0  # should save to disk
        self.sentNotifications = 0
        # Notify bits, they control whether the daemon sends notifications.
        # these are saved to disk
        self.notifyModeChange = False  # 1
        self.notifyStratumChange = False  # 2
        self.notifySyspeerChange = False  # 3
        self.notifyAddAssociation = False  # 4
        self.notifyRMAssociation = False  # 5
        self.notifyConfigChange = False  # 6 [This is not implemented]
        self.notifyLeapSecondAnnounced = False  # 7
        self.notifyHeartbeat = False  # 8
        self.misc_loadDynamicSettings()

    # =============================================================
    # Data read callbacks start here
    # comment divider lines represent not yet implemented callbacks
    # =============================================================

    # Blank: notification OIDs

    def cbr_systemInfo(self, oid, category=None):
        if category == "name":  # The product name of the running NTP
            data = "NTPsec"
        elif category == "version":  # version string
            data = ntp.util.stdversion()
        elif category == "vendor":  # vendor/author name
            data = "Internet Civil Engineering Institute"
        elif category == "system":  # system / hardware info
            # Extract sysname, release, machine from os.uname() tuple
            uname = os.uname()
            data = " ".join([uname[0], uname[2], uname[4]])
        vb = ax.Varbind(ax.VALUE_OCTET_STR, oid, data)
        return vb

    def cbr_timeResolution(self, oid):
        # Uinteger32
        # Arrives in fractional milliseconds
        fuzz = self.safeReadvar(0, ["fuzz"])
        if fuzz is None:
            return None
        fuzz = fuzz["fuzz"]
        # We want to emit fractions of seconds
        # Yes we are flooring instead of rounding: don't want to emit a
        # resolution value higher than ntpd actually produces.
        if fuzz != 0:
            fuzz = int(1 / fuzz)
        else:
            fuzz = 0
        return ax.Varbind(ax.VALUE_GAUGE32, oid, fuzz)

    def cbr_timePrecision(self, oid):
        return self.readCallbackSkeletonSimple(oid, "precision",
                                               ax.VALUE_INTEGER)

    def cbr_timeDistance(self, oid):
        # Displaystring
        data = self.safeReadvar(0, ["rootdist"], raw=True)
        if data is None:
            return None
        data = ntp.util.unitifyvar(data["rootdist"][1], "rootdist",
                                   width=None, unitSpace=True)
        return ax.Varbind(ax.VALUE_OCTET_STR, oid, data)

    # Blank: ntpEntStatus

    def cbr_statusCurrentMode(self, oid):
        mode = self.misc_getMode()
        return ax.Varbind(ax.VALUE_INTEGER, oid, mode)

    def cbr_statusStratum(self, oid):
        # NTPstratum
        return self.readCallbackSkeletonSimple(oid, "stratum",
                                               ax.VALUE_GAUGE32)

    def cbr_statusActiveRefSourceID(self, oid):
        # range of uint32
        syspeer = self.misc_getSyspeerID()
        return ax.Varbind(ax.VALUE_GAUGE32, oid, syspeer)

    def cbr_statusActiveRefSourceName(self, oid):
        # utf8
        data = self.safeReadvar(0, ["peeradr"])
        if data is None:
            return None
        data = ntp.util.canonicalize_dns(data["peeradr"])
        return ax.Varbind(ax.VALUE_OCTET_STR, oid, data)

    def cbr_statusActiveOffset(self, oid):
        # DisplayString
        data = self.safeReadvar(0, ["koffset"], raw=True)
        if data is None:
            return None
        data = ntp.util.unitifyvar(data["koffset"][1], "koffset",
                                   width=None, unitSpace=True)
        return ax.Varbind(ax.VALUE_OCTET_STR, oid, data)

    def cbr_statusNumRefSources(self, oid):
        # range of uint32
        try:
            data = self.session.readstat()
            return ax.Varbind(ax.VALUE_GAUGE32, oid, len(data))
        except ntp.packet.ControlException:
            return None

    def cbr_statusDispersion(self, oid):
        # DisplayString
        data = self.safeReadvar(0, ["rootdisp"], raw=True)
        if data is None:
            return None
        return ax.Varbind(ax.VALUE_OCTET_STR, oid, data["rootdisp"][1])

    def cbr_statusEntityUptime(self, oid):
        # TimeTicks
        # What the spec claims:
        #   The uptime of the NTP entity, (i.e., the time since ntpd was
        #   (re-)initialized not sysUptime!).  The time is represented in
        #   hundreds of seconds since Jan 1, 1970 (00:00:00.000) UTC.
        #
        # First problem: TimeTicks represents hundred*ths* of seconds, could
        #  easily be a typo.
        # Second problem: snmpwalk will happily give you a display of
        #  how long a period of time a value is, such as uptime since start.
        #  That is the opposite of what the spec claims.
        #
        # I am abandoning the spec, and going with what makes a lick of sense
        uptime = self.safeReadvar(0, ["ss_reset"])
        if uptime is None:
            return None
        uptime = uptime["ss_reset"] * 100
        return ax.Varbind(ax.VALUE_TIME_TICKS, oid, uptime)

    def cbr_statusDateTime(self, oid):
        # NtpDateTime
        data = self.safeReadvar(0, ["reftime"])
        if data is None:
            return None
        txt = data["reftime"]
        value = ntp.util.deformatNTPTime(txt)
        return ax.Varbind(ax.VALUE_OCTET_STR, oid, value)

    def cbr_statusLeapSecond(self, oid):  # I am not confident in this yet
        # NtpDateTime
        DAY = 86400
        fmt = "%.8x%.8x"
        data = self.safeReadvar(0, ["reftime"])
        hasleap = self.safeReadvar(0, ["leap"])
        if (data is None) or (hasleap is None):
            return None
        data = data["reftime"]
        hasleap = hasleap["leap"]
        if hasleap in (1, 2):
            seconds = int(data.split(".")[0], 0)
            days = seconds // DAY
            scheduled = (days * DAY) + (DAY - 1)  # 23:59:59 of $CURRENT_DAY
            formatted = fmt % (scheduled, 0)
        else:
            formatted = fmt % (0, 0)
        value = ntp.util.hexstr2octets(formatted)
        return ax.Varbind(ax.VALUE_OCTET_STR, oid, value)

    def cbr_statusLeapSecDirection(self, oid):
        # range of int32
        leap = self.safeReadvar(0, ["leap"])
        if leap is None:
            return None
        leap = leap["leap"]
        if leap == 1:
            pass  # leap 1 == forward
        elif leap == 2:
            leap = -1  # leap 2 == backward
        else:
            leap = 0  # leap 0 or 3 == no change
        return ax.Varbind(ax.VALUE_INTEGER, oid, leap)

    def cbr_statusInPkts(self, oid):
        return self.readCallbackSkeletonSimple(oid, "io_received",
                                               ax.VALUE_COUNTER32)

    def cbr_statusOutPkts(self, oid):
        return self.readCallbackSkeletonSimple(oid, "io_sent",
                                               ax.VALUE_COUNTER32)

    def cbr_statusBadVersion(self, oid):
        return self.readCallbackSkeletonSimple(oid, "ss_oldver",
                                               ax.VALUE_COUNTER32)

    def cbr_statusProtocolError(self, oid):
        data = self.safeReadvar(0, ["ss_badformat", "ss_badauth"])
        if data is None:
            return None
        protoerr = 0
        for key in data.keys():
            protoerr += data[key]
        return ax.Varbind(ax.VALUE_COUNTER32, oid, protoerr)

    def cbr_statusNotifications(self, oid):
        return ax.Varbind(ax.VALUE_COUNTER32, oid, self.sentNotifications)

    ##############################

    # == Dynamics ==
    # assocID
    # assocName
    # assocRefID
    # assocAddrType
    # assocAddr
    # assocOffset
    # assocStratum
    # assocJitter
    # assocDelay
    # assocDispersion
    # assocInPackets
    # assocOutPackets
    # assocProtocolErrors

    #########################

    def cbr_entHeartbeatInterval(self, oid):
        # uint32
        return ax.Varbind(ax.VALUE_GAUGE32, oid, self.heartbeatInterval)

    def cbr_entNotifBits(self, oid):
        # BITS
        data = ax.bools2Bits((False,  # notUsed(0)
                              self.notifyModeChange,
                              self.notifyStratumChange,
                              self.notifySyspeerChange,
                              self.notifyAddAssociation,
                              self.notifyRMAssociation,
                              self.notifyConfigChange,
                              self.notifyLeapSecondAnnounced,
                              self.notifyHeartbeat))
        return ax.Varbind(ax.VALUE_OCTET_STR, oid, data)

    ##########################

    def cbr_entNotifMessage(self, oid):
        # utf8str
        return ax.Varbind(ax.VALUE_OCTET_STR, oid, "no event")

    #########################

    # =====================================
    # Data write callbacks
    # Returns an error value (or noError)
    # Must check that the value is correct for the bind, this does not mean
    #  the type: the master agent handles that
    # Actions: test, undo, commit, cleanup
    # =====================================

    def cbw_entHeartbeatInterval(self, action, varbind, oldData=None):
        if action == "test":
            return ax.ERR_NOERROR
        elif action == "commit":
            self.heartbeatInterval = varbind.payload
            self.misc_storeDynamicSettings()
            return ax.ERR_NOERROR
        elif action == "undo":
            self.heartbeatInterval = oldData
            self.misc_storeDynamicSettings()
            return ax.ERR_NOERROR
        elif action == "cleanup":
            pass

    def cbw_entNotifBits(self, action, varbind, oldData=None):
        if action == "test":
            return ax.ERR_NOERROR
        elif action == "commit":
            (self.notifyModeChange,
             self.notifyStratumChange,
             self.notifySyspeerChange,
             self.notifyAddAssociation,
             self.notifyRMAssociation,
             self.notifyConfigChange,
             self.notifyLeapSecondAnnounced,
             self.notifyHeartbeat) = ax.bits2Bools(varbind.payload, 8)
            self.misc_storeDynamicSettings()
            return ax.ERR_NOERROR
        elif action == "undo":
            (self.notifyModeChange,
             self.notifyStratumChange,
             self.notifySyspeerChange,
             self.notifyAddAssociation,
             self.notifyRMAssociation,
             self.notifyConfigChange,
             self.notifyLeapSecondAnnounced,
             self.notifyHeartbeat) = ax.bits2Bools(oldData, 8)
            self.misc_storeDynamicSettings()
            return ax.ERR_NOERROR
        elif action == "cleanup":
            pass

    # ========================================================================
    # Dynamic tree generator callbacks
    #
    # The structure of these callbacks is somewhat complicated because they
    # share code that is potentially finicky.
    #
    # The dynamicCallbackSkeleton() method handles the construction of the
    # MIB tree, and the placement of the handler() within it. It also provides
    # some useful data to the handler() via the readCallback() layer.
    # ========================================================================

    # Packet Mode Table
    # These are left as stubs for now. Information is lacking on where the
    # data should come from.

    def sub_statPktMode(self):
        pass

    def sub_statPktSent(self):
        pass

    def sub_statPktRecv(self):
        pass

    # Association Table

    def sub_assocID(self):
        def handler(oid, associd):
            return ax.Varbind(ax.VALUE_GAUGE32, oid, associd)
        return self.dynamicCallbackSkeleton(handler)

    def sub_assocName(self):
        return self.dynamicCallbackPeerdata("srcadr", True,
                                            ax.VALUE_OCTET_STR)

    def sub_assocRefID(self):
        def handler(oid, associd):
            pdata = self.misc_getPeerData()
            if pdata is None:
                return None
            # elaborate code in util.py indicates this may not be stable
            try:
                refid = pdata[associd]["refid"][1]
            except IndexError:
                refid = ""
            return ax.Varbind(ax.VALUE_OCTET_STR, oid, refid)
        return self.dynamicCallbackSkeleton(handler)

    def sub_assocAddrType(self):
        def handler(oid, associd):
            pdata = self.misc_getPeerData()
            if pdata is None:
                return None
            srcadr = pdata[associd]["srcadr"][1]
            try:
                socklen = len(socket.getaddrinfo(srcadr, None)[0][-1])
            except socket.gaierror:
                socklen = None
            if socklen == 2:  # ipv4
                addrtype = 1
            elif socklen == 4:  # ipv6
                addrtype = 2
            else:
                # there is also ipv4z and ipv6z..... don't know how to
                # detect those yet. Or if I even need to.
                addrtype = 0  # is this ok? or should it return a NULL?
            return ax.Varbind(ax.VALUE_INTEGER, oid, addrtype)
        return self.dynamicCallbackSkeleton(handler)

    def sub_assocAddr(self):
        def handler(oid, associd):
            pdata = self.misc_getPeerData()
            if pdata is None:
                return None
            srcadr = pdata[associd]["srcadr"][1]
            # WARNING: I am only guessing that this is correct
            # Discover what type of address we have
            try:
                sockinfo = socket.getaddrinfo(srcadr, None)[0][-1]
                addr = sockinfo[0]
                ipv6 = True if len(sockinfo) == 4 else False
            except socket.gaierror:
                addr = None  # how to handle?
                ipv6 = None
            # Convert address string to octets
            srcadr = []
            if not ipv6:
                pieces = addr.split(".")
                for piece in pieces:
                    try:
                        srcadr.append(int(piece))  # feed it a list of ints
                    except ValueError:
                        # Have gotten piece == "" before. Skip over that.
                        # Still try to return data because it is potential
                        # debugging information.
                        continue
            elif ipv6:
                pieces = addr.split(":")
                for piece in pieces:
                    srcadr.append(ntp.util.hexstr2octets(piece))
                srcadr = "".join(srcadr)  # feed it an octet string
            # The octet string encoder can handle either chars or 0-255
            # ints. We use both of those options.
            return ax.Varbind(ax.VALUE_OCTET_STR, oid, srcadr)
        return self.dynamicCallbackSkeleton(handler)

    def sub_assocOffset(self):
        def handler(oid, associd):
            pdata = self.misc_getPeerData()
            if pdata is None:
                return None
            offset = pdata[associd]["offset"][1]
            offset = ntp.util.unitifyvar(offset, "offset", width=None,
                                         unitSpace=True)
            return ax.Varbind(ax.VALUE_OCTET_STR, oid, offset)
        return self.dynamicCallbackSkeleton(handler)

    def sub_assocStratum(self):
        return self.dynamicCallbackPeerdata("stratum", False,
                                            ax.VALUE_GAUGE32)

    def sub_assocJitter(self):
        return self.dynamicCallbackPeerdata("jitter", True,
                                            ax.VALUE_OCTET_STR)

    def sub_assocDelay(self):
        return self.dynamicCallbackPeerdata("delay", True,
                                            ax.VALUE_OCTET_STR)

    def sub_assocDispersion(self):
        return self.dynamicCallbackPeerdata("rootdisp", True,
                                            ax.VALUE_OCTET_STR)

    def sub_assocStatInPkts(self):
        def handler(oid, associd):
            inpkts = self.safeReadvar(associd, ["received"])
            if inpkts is None:
                return None
            inpkts = inpkts["received"]
            return ax.Varbind(ax.VALUE_COUNTER32, oid, inpkts)
        return self.dynamicCallbackSkeleton(handler)

    def sub_assocStatOutPkts(self):
        def handler(oid, associd):
            outpkts = self.safeReadvar(associd, ["sent"])
            if outpkts is None:
                return None
            outpkts = outpkts["sent"]
            return ax.Varbind(ax.VALUE_COUNTER32, oid, outpkts)
        return self.dynamicCallbackSkeleton(handler)

    def sub_assocStatProtoErr(self):
        def handler(oid, associd):
            pvars = self.safeReadvar(associd, ["badauth", "bogusorg",
                                               "seldisp", "selbroken"])
            if pvars is None:
                return None
            protoerr = 0
            for key in pvars.keys():
                protoerr += pvars[key]
            return ax.Varbind(ax.VALUE_COUNTER32, oid, protoerr)
        return self.dynamicCallbackSkeleton(handler)

    # =====================================
    # Notification handlers
    # =====================================

    def checkNotifications(self, control):
        currentTime = time.time()
        if (currentTime - self.lastNotifyCheck) < self.notifySpinTime:
            return
        self.lastNotifyCheck = currentTime

        if self.notifyModeChange:
            self.doNotifyModeChange(control)

        if self.notifyStratumChange:
            self.doNotifyStratumChange(control)

        if self.notifySyspeerChange:
            self.doNotifySyspeerChange(control)

        # Both add and remove have to look at the same data, don't want them
        # stepping on each other. Therefore the functions are combined.
        if self.notifyAddAssociation and self.notifyRMAssociation:
            self.doNotifyChangeAssociation(control, "both")
        elif self.notifyAddAssociation:
            self.doNotifyChangeAssociation(control, "add")
        elif self.notifyRMAssociation:
            self.doNotifyChangeAssociation(control, "rm")

        if self.notifyConfigChange:
            self.doNotifyConfigChange(control)

        if self.notifyLeapSecondAnnounced:
            self.doNotifyLeapSecondAnnounced(control)

        if self.notifyHeartbeat:
            self.doNotifyHeartbeat(control)

    def doNotifyModeChange(self, control):
        oldMode = self.oldValues.get("mode")
        newMode = self.misc_getMode()  # connection failure handled by method
        if oldMode is None:
            self.oldValues["mode"] = newMode
            return
        elif oldMode != newMode:
            self.oldValues["mode"] = newMode
            vl = [ax.Varbind(ax.VALUE_OID, snmpTrapOID,
                             ax.OID(ntpRootOID + (0, 1))),
                  ax.Varbind(ax.VALUE_INTEGER, ntpRootOID + (1, 2, 1),
                             newMode)]
            control.sendNotify(vl)
            self.sentNotifications += 1

    def doNotifyStratumChange(self, control):
        oldStratum = self.oldValues.get("stratum")
        newStratum = self.safeReadvar(0, ["stratum"])
        if newStratum is None:
            return  # couldn't read
        newStratum = newStratum["stratum"]
        if oldStratum is None:
            self.oldValues["stratum"] = newStratum
            return
        elif oldStratum != newStratum:
            self.oldValues["stratum"] = newStratum
            datetime = self.safeReadvar(0, ["reftime"])
            if datetime is None:
                datetime = ""
            else:
                datetime = ntp.util.deformatNTPTime(datetime["reftime"])
            vl = [ax.Varbind(ax.VALUE_OID, snmpTrapOID,
                             ax.OID(ntpRootOID + (0, 2))),
                  ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 2, 9),
                             datetime),
                  ax.Varbind(ax.VALUE_GAUGE32, ntpRootOID + (1, 2, 2),
                             newStratum),
                  ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 5, 1),
                             "Stratum changed")]  # Uh... what goes here?
            control.sendNotify(vl)
            self.sentNotifications += 1

    def doNotifySyspeerChange(self, control):
        oldSyspeer = self.oldValues.get("syspeer")
        newSyspeer = self.safeReadvar(0, ["peeradr"])
        if newSyspeer is None:
            return  # couldn't read
        newSyspeer = newSyspeer["peeradr"]
        if oldSyspeer is None:
            self.oldValues["syspeer"] = newSyspeer
            return
        elif oldSyspeer != newSyspeer:
            self.oldValues["syspeer"] = newSyspeer
            datetime = self.safeReadvar(0, ["reftime"])
            if datetime is None:
                datetime = ""
            else:
                datetime = ntp.util.deformatNTPTime(datetime["reftime"])
            syspeer = self.misc_getSyspeerID()
            vl = [ax.Varbind(ax.VALUE_OID, snmpTrapOID,
                             ax.OID(ntpRootOID + (0, 3))),
                  ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 2, 9),
                             datetime),
                  ax.Varbind(ax.VALUE_GAUGE32, ntpRootOID + (1, 2, 3),
                             syspeer),
                  ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 5, 1),
                             "SysPeer changed")]  # Uh... what goes here?
            control.sendNotify(vl)
            self.sentNotifications += 1

    def doNotifyChangeAssociation(self, control, which):
        # Add and remove are combined because they use the same data source
        # and it would be easy to have them stepping on each other.
        changes = self.misc_getAssocListChanges()
        if changes is None:
            return
        datetime = self.safeReadvar(0, ["reftime"])
        if datetime is None:
            datetime = ""
        else:
            datetime = ntp.util.deformatNTPTime(datetime["reftime"])
        adds, rms = changes
        if which in ("add", "both"):
            for name in adds:
                vl = [ax.Varbind(ax.VALUE_OID, snmpTrapOID,
                                 ax.OID(ntpRootOID + (0, 4))),  # Add
                      ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 2, 9),
                                 datetime),
                      ax.Varbind(ax.VALUE_OCTET_STR,
                                 ntpRootOID + (1, 3, 1, 1, 2),
                                 name),
                      ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 5, 1),
                                 "Association added")]
                control.sendNotify(vl)
                self.sentNotifications += 1
        if which in ("rm", "both"):
            for name in rms:
                vl = [ax.Varbind(ax.VALUE_OID, snmpTrapOID,
                                 ax.OID(ntpRootOID + (0, 5))),  # Remove
                      ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 2, 9),
                                 datetime),
                      ax.Varbind(ax.VALUE_OCTET_STR,
                                 ntpRootOID + (1, 3, 1, 1, 2),
                                 name),
                      ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 5, 1),
                                 "Association removed")]
                control.sendNotify(vl)
                self.sentNotifications += 1

    def doNotifyConfigChange(self, control):
        # This left unimplemented because the MIB wants something we can't
        # and/or shouldn't provide
        pass

    def doNotifyLeapSecondAnnounced(self, control):
        oldLeap = self.oldValues.get("leap")
        newLeap = self.safeReadvar(0, ["leap"])
        if newLeap is None:
            return
        newLeap = newLeap["leap"]
        if oldLeap is None:
            self.oldValues["leap"] = newLeap
            return
        if oldLeap != newLeap:
            self.oldValues["leap"] = newLeap
            if (oldLeap in (0, 3)) and (newLeap in (1, 2)):
                # changed noleap or unsync to a leap announcement
                datetime = self.safeReadvar(0, ["reftime"])
                if datetime is None:
                    datetime = ""
                else:
                    datetime = ntp.util.deformatNTPTime(datetime["reftime"])
                vl = [ax.Varbind(ax.VALUE_OID, snmpTrapOID,
                                 ax.OID(ntpRootOID + (0, 7))),
                      ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 2, 9),
                                 datetime),
                      ax.Varbind(ax.VALUE_OCTET_STR, ntpRootOID + (1, 5, 1),
                                 "Leap second announced")]
                control.sendNotify(vl)
                self.sentNotifications += 1

    def doNotifyHeartbeat(self, control):  # TODO: check if ntpd running?
        vl = [ax.Varbind(ax.VALUE_OID, snmpTrapOID,
                         ax.OID(ntpRootOID + (0, 8))),
              ax.Varbind(ax.VALUE_GAUGE32, ntpRootOID + (0, 1, 4, 1),
                         self.heartbeatInterval)]
        if self.heartbeatInterval == 0:  # interval == 0 means send once
            self.notifyHeartbeat = False
            control.sendNotify(vl)
            self.sentNotifications += 1
        else:
            current = ntp.util.monoclock()
            if (current - self.lastHeartbeat) > self.heartbeatInterval:
                self.lastHeartbeat = current
                control.sendNotify(vl)
                self.sentNotifications += 1

    # =====================================
    # Misc data helpers (not part of the MIB proper)
    # =====================================

    def misc_loadDynamicSettings(self):
        if self.settingsFilename is None:
            return

        def boolify(d, k):
            return True if d[k][0][1] == "True" else False
        optionList = ("notify-mode-change", "notify-stratum-change",
                      "notify-syspeer-change", "notify-add-association",
                      "notify-rm-association", "notify-leap-announced",
                      "notify-heartbeat", "heartbeat-interval")
        settings = loadSettings(self.settingsFilename, optionList)
        if settings is None:
            return
        for key in settings.keys():
            if key == "notify-mode-change":
                self.notifyModeChange = boolify(settings, key)
            elif key == "notify-stratum-change":
                self.notifyStratumChange = boolify(settings, key)
            elif key == "notify-syspeer-change":
                self.notifySyspeerChange = boolify(settings, key)
            elif key == "notify-add-association":
                self.notifyAddAssociation = boolify(settings, key)
            elif key == "notify-rm-association":
                self.notifyRMAssociation = boolify(settings, key)
            elif key == "notify-leap-announced":
                self.notifyLeapSecondAnnounced = boolify(settings, key)
            elif key == "notify-heartbeat":
                self.notifyHeartbeat = boolify(settings, key)
            elif key == "heartbeat-interval":
                self.heartbeatInterval = settings[key][0][1]

    def misc_storeDynamicSettings(self):
        if self.settingsFilename is None:
            return
        settings = {}
        settings["notify-mode-change"] = str(self.notifyModeChange)
        settings["notify-stratum-change"] = str(self.notifyStratumChange)
        settings["notify-syspeer-change"] = str(self.notifySyspeerChange)
        settings["notify-add-association"] = str(self.notifyAddAssociation)
        settings["notify-rm-association"] = str(self.notifyRMAssociation)
        settings["notify-leap-announced"] = str(self.notifyLeapSecondAnnounced)
        settings["notify-heartbeat"] = str(self.notifyHeartbeat)
        settings["heartbeat-interval"] = str(self.heartbeatInterval)
        storeSettings(self.settingsFilename, settings)

    def misc_getAssocListChanges(self):
        # We need to keep the names, because those won't be available
        # after they have been removed.
        oldAssoc = self.oldValues.get("assoc")
        newAssoc = {}
        # Yes, these are cached, for a very short time
        pdata = self.misc_getPeerData()
        if pdata is None:
            return
        ids = self.misc_getPeerIDs()
        if ids is None:
            return
        for associd in ids:
            addr = pdata[associd]["srcadr"][1]
            name = ntp.util.canonicalize_dns(addr)
            newAssoc[associd] = name
        if oldAssoc is None:
            self.oldValues["assoc"] = newAssoc
            return
        elif oldAssoc != newAssoc:
            oldIDs = oldAssoc.keys()
            newIDs = newAssoc.keys()
            adds = []
            rms = []
            for associd in oldIDs + newIDs:
                if associd not in newIDs:  # removed
                    rms.append(oldAssoc[associd])
                if associd not in oldIDs:  # added
                    adds.append(newAssoc[associd])
            return (adds, rms)
        return

    def misc_getMode(self):  # FIXME: not fully implemented
        try:
            # Don't care about the data, this is a ploy to get the rstatus
            self.session.readvar(0, ["stratum"])
        except ntp.packet.ControlException as e:
            if e.message == ntp.packet.SERR_SOCKET:
                # Can't connect, ntpd probably not running
                return 1
            else:
                raise e
        rstatus = self.session.rstatus  # a ploy to get the system status
        source = ntp.control.CTL_SYS_SOURCE(rstatus)
        if source == ntp.control.CTL_SST_TS_UNSPEC:
            mode = 2  # Not yet synced
        elif False:
            mode = 3  # No reference configured
        elif source == ntp.control.CTL_SST_TS_LOCAL:
            mode = 4  # Distributing local clock (low accuracy)
        elif source in (ntp.control.CTL_SST_TS_ATOM,
                        ntp.control.CTL_SST_TS_LF,
                        ntp.control.CTL_SST_TS_HF,
                        ntp.control.CTL_SST_TS_UHF):
            # I am not sure if I should be including the radios in this
            mode = 5  # Synced to local refclock
        elif source == ntp.control.CTL_SST_TS_NTP:
            # Should this include "other"? That covers things like chrony...
            mode = 6  # Sync to remote NTP
        else:
            mode = 99  # Unknown
        return mode

    def misc_getSyspeerID(self):
        peers = self.misc_getPeerData()
        syspeer = 0
        for associd in peers.keys():
            rstatus = peers[associd]["peerstatus"]
            if (ntp.control.CTL_PEER_STATVAL(rstatus) & 0x7) == \
               ntp.control.CTL_PST_SEL_SYSPEER:
                syspeer = associd
                break
        return syspeer

    def safeReadvar(self, associd, variables=None, raw=False):
        # Use this when we want to catch packet errors, but don't care
        # about what they are
        try:
            return self.session.readvar(associd, varlist=variables, raw=raw)
        except ntp.packet.ControlException:
            return None

    def dynamicCallbackPeerdata(self, variable, raw, valueType):
        rawindex = 1 if raw else 0

        def handler(oid, associd):
            pdata = self.misc_getPeerData()
            if pdata is None:
                return None
            value = pdata[associd][variable][rawindex]
            return ax.Varbind(valueType, oid, value)
        return self.dynamicCallbackSkeleton(handler)

    def dynamicCallbackSkeleton(self, handler):
        # Build a dynamic MIB tree, installing the provided handler in it
        def readCallback(oid):
            # This function assumes that it is a leaf node and that the
            # last number in the OID is the index.
            index = oid.subids[-1]  # if called properly this works (Ha!)
            index -= 1  # SNMP reserves index 0, effectively 1-based lists
            associd = self.misc_getPeerIDs()[index]
            return handler(oid, associd)
        subs = {}
        associds = self.misc_getPeerIDs()  # need the peer count
        for i in range(len(associds)):
            subs[i+1] = {"reader": readCallback}
        return subs

    def readCallbackSkeletonSimple(self, oid, varname, dataType):
        # Used for entries that just need a simple variable retrevial
        # but do not need any processing.
        data = self.safeReadvar(0, [varname])
        if data is None:
            return None
        else:
            return ax.Varbind(dataType, oid, data[varname])

    def misc_getPeerIDs(self):
        peerids = self.cache.get("peerids")
        if peerids is None:
            try:
                peerids = [x.associd for x in self.session.readstat()]
            except ntp.packet.ControlException:
                peerids = []
            peerids.sort()
            self.cache.set("peerids", peerids)
        return peerids

    def misc_getPeerData(self):
        peerdata = self.cache.get("peerdata")
        if peerdata is None:
            associds = self.misc_getPeerIDs()
            peerdata = {}
            for aid in associds:
                try:
                    pdata = self.safeReadvar(aid, raw=True)
                    pdata["peerstatus"] = self.session.rstatus
                except IOError:
                    continue
                peerdata[aid] = pdata
            self.cache.set("peerdata", peerdata)
        return peerdata


def connect(address):
    try:
        if type(address) is str:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect(address)
        else:
            host, port = address[0], address[1]
            af, _, _, _, _ = socket.getaddrinfo(host, port)[0]
            sock = socket.socket(af, socket.SOCK_STREAM)
            sock.connect((host, port))
    except socket.error as msg:
        log("Connection to %s failure: %s" % (repr(address), repr(msg)), 1)
        sys.exit(1)
    log("connected to master agent at " + repr(address), 3)
    return sock


def mainloop(snmpSocket, reconnectionAddr, host=None):
    log("initing loop", 3)
    dbase = DataSource(host, "/var/ntpsntpd/notify.conf")
    while True:  # Loop reconnection attempts
        control = PacketControl(snmpSocket, dbase, logfp=logfp, debug=debug)
        control.loopCallback = dbase.checkNotifications
        control.initNewSession()
        if not control.mainloop(True):  # disconnected
            snmpSocket.close()
            snmpSocket = connect(reconnectionAddr)
            log("disconnected from master, attempting reconnect", 2)
        else:  # Something else happened
            break


def daemonize(runfunc, *runArgs):
    pid = os.fork()
    if pid < 0:
        log("Forking error " + str(pid), 1)
        sys.exit(pid)
    elif pid > 0:  # We are the parent
        log("Daemonization success, child pid: " + str(pid), 3)
        sys.exit(0)

    # We must be the child

    os.umask(0)

    sid = os.setsid()

    # chdir should be here, change to what? root?

    global logfp
    if logfp == sys.stderr:
        logfp = None

    sys.stdin.close()
    sys.stdin = None
    sys.stdout.close()
    sys.stdout = None
    sys.stderr.close()
    sys.stderr = None

    runfunc(*runArgs)


def loadSettings(filename, optionList):
    log("Loading config file: %s" % filename, 3)
    if not os.path.isfile(filename):
        return None
    options = {}
    with open(filename) as f:
        data = f.read()
        lines = ntp.util.parseConf(data)
        for line in lines:
            isQuote, token = line[0]
            if token in optionList:
                options[token] = line[1:]
    return options


def storeSettings(filename, settings):
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    data = []
    for key in settings.keys():
        data.append("%s %s\n" % (key, settings[key]))
    data = "".join(data)
    with open(filename, "w") as f:
        f.write(data)


usage = """
USAGE: ntpsnmpd [-n] [ntp host]
  Flg Arg Option-Name   Description
   -n no  no-fork         Do not fork and daemonize.
   -x Adr master-addr     Specify address for connecting to the master agent
                                - default /var/agentx/master
   -d no  debug-level     Increase output debug message level
                                - may appear multiple times
   -l Str logfile         Logs debug messages to the provided filename
   -D Int set-debug-level Set the output debug message level
                                - may appear multiple times
   -h no  help            Print a usage message.
   -V no  version         Output version information and exit
"""


if __name__ == "__main__":
    bin_ver = "ntpsec-@NTPSEC_VERSION_EXTENDED@"
    ntp.util.stdversioncheck(bin_ver)
    try:
        (options, arguments) = getopt.getopt(
            sys.argv[1:],
            "nx:dD:Vhl:c:",
            ["no-fork", "master-address=", "debug-level", "set-debug-level=",
             "version", "help", "logfile=", "configfile="])
    except getopt.GetoptError as e:
        sys.stderr.write("%s\n" % e)
        sys.stderr.write(usage)
        raise SystemExit(1)

    masterAddr = "/var/agentx/master"
    logfile = DEFLOG
    hostname = DEFHOST

    # Check for non-default config-file
    conffile = "/etc/ntpsnmpd.conf"
    for (switch, val) in options:
        if switch in ("-c", "--configfile"):
            conffile = val
            break

    # Load configuration file
    conf = loadSettings(conffile,
                        ("master-addr", "logfile", "loglevel", "ntp-addr"))
    if conf is not None:
        for key in conf.keys():
            if key == "master-addr":  # Address of the SNMP master daemon
                val = conf[key][0][1]
                if ":" in val:
                    host, port = val.split(":")
                    port = int(port)
                    masterAddr = (host, port)
                else:
                    masterAddr = val
            elif key == "logfile":
                logfile = conf[key][0][1]
            elif key == "ntp-addr":  # Address of the NTP daemon
                hostname = conf[key][0][1]
            elif key == "loglevel":
                errmsg = "Error: loglevel parameter '%s' not a number\n"
                debug = conf[key][0][1]

    fileLogging = False
    for (switch, val) in options:
        if switch in ("-n", "--no-fork"):
            nofork = True
        elif switch in ("-x", "--master-addr"):
            if ":" in val:
                host, port = val.split(":")
                port = int(port)
                masterAddr = (host, port)
            else:
                masterAddr = val
        elif switch in ("-d", "--debug-level"):
            debug += 1
        elif switch in ("-D", "--set-debug-level"):
            errmsg = "Error: -D parameter '%s' not a number\n"
            debug = ntp.util.safeargcast(val, int, errmsg, usage)
        elif switch in ("-V", "--version"):
            print("ntpsnmpd %s" % ntp.util.stdversion())
            raise SystemExit(0)
        elif switch in ("-h", "--help"):
            print(usage)
            raise SystemExit(0)
        elif switch in ("-l", "--logfile"):
            logfile = val
            fileLogging = True

    if not nofork:
        fileLogging = True

    if fileLogging:
        if logfp != sys.stderr:
            logfp.close()
        logfp = open(logfile, "a", 1)  # 1 => line buffered

    hostname = arguments[0] if arguments else DEFHOST

    # Connect here so it can always report a connection error
    sock = connect(masterAddr)

    if nofork:
        mainloop(sock, hostname)
    else:
        daemonize(mainloop, sock, hostname)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import atexit,base64,errno,getpass,os,re,shutil,sys,time,threading,traceback,shlex
try:
	import subprocess32 as subprocess
except ImportError:
	import subprocess
base_cache=os.path.expanduser('~/.cache/')
if not os.path.isdir(base_cache):
	base_cache='/tmp/'
default_wafcache_dir=os.path.join(base_cache,'wafcache_'+getpass.getuser())
CACHE_DIR=os.environ.get('WAFCACHE',default_wafcache_dir)
WAFCACHE_CMD=os.environ.get('WAFCACHE_CMD')
TRIM_MAX_FOLDERS=int(os.environ.get('WAFCACHE_TRIM_MAX_FOLDER',1000000))
EVICT_INTERVAL_MINUTES=int(os.environ.get('WAFCACHE_EVICT_INTERVAL_MINUTES',3))
EVICT_MAX_BYTES=int(os.environ.get('WAFCACHE_EVICT_MAX_BYTES',10**10))
WAFCACHE_NO_PUSH=1 if os.environ.get('WAFCACHE_NO_PUSH')else 0
WAFCACHE_VERBOSITY=1 if os.environ.get('WAFCACHE_VERBOSITY')else 0
WAFCACHE_STATS=1 if os.environ.get('WAFCACHE_STATS')else 0
WAFCACHE_ASYNC_WORKERS=os.environ.get('WAFCACHE_ASYNC_WORKERS')
WAFCACHE_ASYNC_NOWAIT=os.environ.get('WAFCACHE_ASYNC_NOWAIT')
OK="ok"
re_waf_cmd=re.compile('(?P<src>%{SRC})|(?P<tgt>%{TGT})')
try:
	import cPickle
except ImportError:
	import pickle as cPickle
if __name__!='__main__':
	from waflib import Task,Logs,Utils,Build
def can_retrieve_cache(self):
	if not self.outputs:
		return False
	self.cached=False
	sig=self.signature()
	ssig=Utils.to_hex(self.uid()+sig)
	if WAFCACHE_STATS:
		self.generator.bld.cache_reqs+=1
	files_to=[node.abspath()for node in self.outputs]
	proc=get_process()
	err=cache_command(proc,ssig,[],files_to)
	process_pool.append(proc)
	if err.startswith(OK):
		if WAFCACHE_VERBOSITY:
			Logs.pprint('CYAN','  Fetched %r from cache'%files_to)
		else:
			Logs.debug('wafcache: fetched %r from cache',files_to)
		if WAFCACHE_STATS:
			self.generator.bld.cache_hits+=1
	else:
		if WAFCACHE_VERBOSITY:
			Logs.pprint('YELLOW','  No cache entry %s'%files_to)
		else:
			Logs.debug('wafcache: No cache entry %s: %s',files_to,err)
		return False
	self.cached=True
	return True
def put_files_cache(self):
	if WAFCACHE_NO_PUSH or getattr(self,'cached',None)or not self.outputs:
		return
	files_from=[]
	for node in self.outputs:
		path=node.abspath()
		if not os.path.isfile(path):
			return
		files_from.append(path)
	bld=self.generator.bld
	old_sig=self.signature()
	for node in self.inputs:
		try:
			del node.ctx.cache_sig[node]
		except KeyError:
			pass
	delattr(self,'cache_sig')
	sig=self.signature()
	def _async_put_files_cache(bld,ssig,files_from):
		proc=get_process()
		if WAFCACHE_ASYNC_WORKERS:
			with bld.wafcache_lock:
				if bld.wafcache_stop:
					process_pool.append(proc)
					return
				bld.wafcache_procs.add(proc)
		err=cache_command(proc,ssig,files_from,[])
		process_pool.append(proc)
		if err.startswith(OK):
			if WAFCACHE_VERBOSITY:
				Logs.pprint('CYAN','  Successfully uploaded %s to cache'%files_from)
			else:
				Logs.debug('wafcache: Successfully uploaded %r to cache',files_from)
			if WAFCACHE_STATS:
				bld.cache_puts+=1
		else:
			if WAFCACHE_VERBOSITY:
				Logs.pprint('RED','  Error caching step results %s: %s'%(files_from,err))
			else:
				Logs.debug('wafcache: Error caching results %s: %s',files_from,err)
	if old_sig==sig:
		ssig=Utils.to_hex(self.uid()+sig)
		if WAFCACHE_ASYNC_WORKERS:
			fut=bld.wafcache_executor.submit(_async_put_files_cache,bld,ssig,files_from)
			bld.wafcache_uploads.append(fut)
		else:
			_async_put_files_cache(bld,ssig,files_from)
	else:
		Logs.debug('wafcache: skipped %r upload due to late input modifications %r',self.outputs,self.inputs)
	bld.task_sigs[self.uid()]=self.cache_sig
def hash_env_vars(self,env,vars_lst):
	if not env.table:
		env=env.parent
		if not env:
			return Utils.SIG_NIL
	idx=str(id(env))+str(vars_lst)
	try:
		cache=self.cache_env
	except AttributeError:
		cache=self.cache_env={}
	else:
		try:
			return self.cache_env[idx]
		except KeyError:
			pass
	v=str([env[a]for a in vars_lst])
	v=v.replace(self.srcnode.abspath().__repr__()[:-1],'')
	m=Utils.md5()
	m.update(v.encode())
	ret=m.digest()
	Logs.debug('envhash: %r %r',ret,v)
	cache[idx]=ret
	return ret
def uid(self):
	try:
		return self.uid_
	except AttributeError:
		m=Utils.md5()
		src=self.generator.bld.srcnode
		up=m.update
		up(self.__class__.__name__.encode())
		for x in self.inputs+self.outputs:
			up(x.path_from(src).encode())
		self.uid_=m.digest()
		return self.uid_
def make_cached(cls):
	if getattr(cls,'nocache',None)or getattr(cls,'has_cache',False):
		return
	full_name="%s.%s"%(cls.__module__,cls.__name__)
	if full_name in('waflib.Tools.ccroot.vnum','waflib.Build.inst'):
		return
	m1=getattr(cls,'run',None)
	def run(self):
		if getattr(self,'nocache',False):
			return m1(self)
		if self.can_retrieve_cache():
			return 0
		return m1(self)
	cls.run=run
	m2=getattr(cls,'post_run',None)
	def post_run(self):
		if getattr(self,'nocache',False):
			return m2(self)
		ret=m2(self)
		self.put_files_cache()
		return ret
	cls.post_run=post_run
	cls.has_cache=True
process_pool=[]
def get_process():
	try:
		return process_pool.pop()
	except IndexError:
		filepath=os.path.dirname(os.path.abspath(__file__))+os.sep+'wafcache.py'
		cmd=[sys.executable,'-c',Utils.readf(filepath)]
		return subprocess.Popen(cmd,stdout=subprocess.PIPE,stdin=subprocess.PIPE,bufsize=0)
def atexit_pool():
	for proc in process_pool:
		proc.kill()
atexit.register(atexit_pool)
def build(bld):
	if WAFCACHE_ASYNC_WORKERS:
		try:
			num_workers=int(WAFCACHE_ASYNC_WORKERS)
		except ValueError:
			Logs.warn('Invalid WAFCACHE_ASYNC_WORKERS specified: %r'%WAFCACHE_ASYNC_WORKERS)
		else:
			from concurrent.futures import ThreadPoolExecutor
			bld.wafcache_executor=ThreadPoolExecutor(max_workers=num_workers)
			bld.wafcache_uploads=[]
			bld.wafcache_procs=set([])
			bld.wafcache_stop=False
			bld.wafcache_lock=threading.Lock()
		def finalize_upload_async(bld):
			if WAFCACHE_ASYNC_NOWAIT:
				with bld.wafcache_lock:
					bld.wafcache_stop=True
				for fut in reversed(bld.wafcache_uploads):
					fut.cancel()
				for proc in bld.wafcache_procs:
					proc.kill()
				bld.wafcache_procs.clear()
			else:
				Logs.pprint('CYAN','... waiting for wafcache uploads to complete (%s uploads)'%len(bld.wafcache_uploads))
			bld.wafcache_executor.shutdown(wait=True)
		bld.add_post_fun(finalize_upload_async)
	if WAFCACHE_STATS:
		bld.cache_reqs=bld.cache_hits=bld.cache_puts=0
		def printstats(bld):
			hit_ratio=0
			if bld.cache_reqs>0:
				hit_ratio=(bld.cache_hits/bld.cache_reqs)*100
			Logs.pprint('CYAN','  wafcache stats: %s requests, %s hits (ratio: %.2f%%), %s writes'%(bld.cache_reqs,bld.cache_hits,hit_ratio,bld.cache_puts))
		bld.add_post_fun(printstats)
	if process_pool:
		return
	processes=[get_process()for x in range(bld.jobs)]
	process_pool.extend(processes)
	Task.Task.can_retrieve_cache=can_retrieve_cache
	Task.Task.put_files_cache=put_files_cache
	Task.Task.uid=uid
	Build.BuildContext.hash_env_vars=hash_env_vars
	for x in reversed(list(Task.classes.values())):
		make_cached(x)
def cache_command(proc,sig,files_from,files_to):
	obj=base64.b64encode(cPickle.dumps([sig,files_from,files_to]))
	proc.stdin.write(obj)
	proc.stdin.write('\n'.encode())
	proc.stdin.flush()
	obj=proc.stdout.readline()
	if not obj:
		raise OSError('Preforked sub-process %r died'%proc.pid)
	return cPickle.loads(base64.b64decode(obj))
try:
	copyfun=os.link
except NameError:
	copyfun=shutil.copy2
def atomic_copy(orig,dest):
	global copyfun
	tmp=dest+'.tmp'
	up=os.path.dirname(dest)
	try:
		os.makedirs(up)
	except OSError:
		pass
	try:
		copyfun(orig,tmp)
	except OSError as e:
		if e.errno==errno.EXDEV:
			copyfun=shutil.copy2
			copyfun(orig,tmp)
		else:
			raise
	os.rename(tmp,dest)
def lru_trim():
	lst=[]
	for up in os.listdir(CACHE_DIR):
		if len(up)==2:
			sub=os.path.join(CACHE_DIR,up)
			for hval in os.listdir(sub):
				path=os.path.join(sub,hval)
				size=0
				for fname in os.listdir(path):
					try:
						size+=os.lstat(os.path.join(path,fname)).st_size
					except OSError:
						pass
				lst.append((os.stat(path).st_mtime,size,path))
	lst.sort(key=lambda x:x[0])
	lst.reverse()
	tot=sum(x[1]for x in lst)
	while tot>EVICT_MAX_BYTES or len(lst)>TRIM_MAX_FOLDERS:
		_,tmp_size,path=lst.pop()
		tot-=tmp_size
		tmp=path+'.remove'
		try:
			shutil.rmtree(tmp)
		except OSError:
			pass
		try:
			os.rename(path,tmp)
		except OSError:
			sys.stderr.write('Could not rename %r to %r\n'%(path,tmp))
		else:
			try:
				shutil.rmtree(tmp)
			except OSError:
				sys.stderr.write('Could not remove %r\n'%tmp)
	sys.stderr.write("Cache trimmed: %r bytes in %r folders left\n"%(tot,len(lst)))
def lru_evict():
	lockfile=os.path.join(CACHE_DIR,'all.lock')
	try:
		st=os.stat(lockfile)
	except EnvironmentError as e:
		if e.errno==errno.ENOENT:
			with open(lockfile,'w')as f:
				f.write('')
		else:
			raise
	if st.st_mtime<time.time()-EVICT_INTERVAL_MINUTES*60:
		fd=os.open(lockfile,os.O_RDWR|os.O_CREAT,0o755)
		try:
			try:
				import fcntl
			except ImportError:
				import msvcrt,ctypes,ctypes.wintypes
				handle=msvcrt.get_osfhandle(fd)
				kernel32=ctypes.windll('kernel32',use_last_error=True)
				DWORD=ctypes.wintypes.DWORD
				HANDLE=ctypes.wintypes.HANDLE
				class DUMMYSTRUCTNAME(ctypes.Structure):
					_fields=[('Offset',ctypes.wintypes.DWORD),('OffsetHigh',DWORD)]
				class DUMMYUNIONNAME(ctypes.Union):
					_fields_=[('_dummystructname',DUMMYSTRUCTNAME),('Pointer',ctypes.c_void_p)]
				class OVERLAPPED(ctypes.Structure):
					_fields_=[('Internal',ctypes.c_void_p),('InternalHigh',ctypes.c_void_p),('_dummyunionname',DUMMYUNIONNAME),('hEvent',HANDLE)]
				LockFileEx=kernel32.LockFileEx
				LockFileEx.argtypes=[HANDLE,DWORD,DWORD,DWORD,DWORD,POINTER(OVERLAPPED)]
				LockFileEx.restype=BOOL
				UnlockFileEx=kernel32.UnlockFileEx
				UnlockFileEx.argtypes=[HANDLE,DWORD,DWORD,DWORD,POINTER(OVERLAPPED)]
				UnlockFileEx.restype=BOOL
				if LockFileEx(handle,3,0,1,0,ctypes.pointer(OVERLAPPED())):
					try:
						lru_trim()
						os.utime(lockfile,None)
					finally:
						win32file.UnlockFileEx(handle,0,1,0,ctypes.pointer(OVERLAPPED()))
				else:
					last_error=kernel32.GetLastError()
					if last_error==33:
						if WAFCACHE_VERBOSITY:
							sys.stderr.write('wafcache: another cleaning process is running\n')
					else:
						raise OSError(last_error)
			else:
				try:
					fcntl.flock(fd,fcntl.LOCK_EX|fcntl.LOCK_NB)
				except EnvironmentError:
					if WAFCACHE_VERBOSITY:
						sys.stderr.write('wafcache: another cleaning process is running\n')
				else:
					lru_trim()
					os.utime(lockfile,None)
		finally:
			os.close(fd)
class netcache(object):
	def __init__(self):
		import urllib3
		self.http=urllib3.PoolManager()
	def url_of(self,sig,i):
		return"%s/%s/%s"%(CACHE_DIR,sig,i)
	def upload(self,file_path,sig,i):
		url=self.url_of(sig,i)
		with open(file_path,'rb')as f:
			file_data=f.read()
		r=self.http.request('POST',url,timeout=60,fields={'file':('%s/%s'%(sig,i),file_data),})
		if r.status>=400:
			raise OSError("Invalid status %r %r"%(url,r.status))
	def download(self,file_path,sig,i):
		url=self.url_of(sig,i)
		with self.http.request('GET',url,preload_content=False,timeout=60)as inf:
			if inf.status>=400:
				raise OSError("Invalid status %r %r"%(url,inf.status))
			with open(file_path,'wb')as out:
				shutil.copyfileobj(inf,out)
	def copy_to_cache(self,sig,files_from,files_to):
		try:
			for i,x in enumerate(files_from):
				if not os.path.islink(x):
					self.upload(x,sig,i)
		except Exception:
			return traceback.format_exc()
		return OK
	def copy_from_cache(self,sig,files_from,files_to):
		try:
			for i,x in enumerate(files_to):
				self.download(x,sig,i)
		except Exception:
			return traceback.format_exc()
		return OK
class fcache(object):
	def __init__(self):
		if not os.path.exists(CACHE_DIR):
			try:
				os.makedirs(CACHE_DIR)
			except OSError:
				pass
		if not os.path.exists(CACHE_DIR):
			raise ValueError('Could not initialize the cache directory')
	def copy_to_cache(self,sig,files_from,files_to):
		try:
			for i,x in enumerate(files_from):
				dest=os.path.join(CACHE_DIR,sig[:2],sig,str(i))
				atomic_copy(x,dest)
		except Exception:
			return traceback.format_exc()
		else:
			try:
				lru_evict()
			except Exception:
				return traceback.format_exc()
		return OK
	def copy_from_cache(self,sig,files_from,files_to):
		try:
			for i,x in enumerate(files_to):
				orig=os.path.join(CACHE_DIR,sig[:2],sig,str(i))
				atomic_copy(orig,x)
			os.utime(os.path.join(CACHE_DIR,sig[:2],sig),None)
		except Exception:
			return traceback.format_exc()
		return OK
class bucket_cache(object):
	def bucket_copy(self,source,target):
		if WAFCACHE_CMD:
			def replacer(match):
				if match.group('src'):
					return source
				elif match.group('tgt'):
					return target
			cmd=[re_waf_cmd.sub(replacer,x)for x in shlex.split(WAFCACHE_CMD)]
		elif CACHE_DIR.startswith('s3://'):
			cmd=['aws','s3','cp',source,target]
		elif CACHE_DIR.startswith('gs://'):
			cmd=['gsutil','cp',source,target]
		else:
			cmd=['mc','cp',source,target]
		proc=subprocess.Popen(cmd,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
		out,err=proc.communicate()
		if proc.returncode:
			raise OSError('Error copy %r to %r using: %r (exit %r):\n  out:%s\n  err:%s'%(source,target,cmd,proc.returncode,out.decode(errors='replace'),err.decode(errors='replace')))
	def copy_to_cache(self,sig,files_from,files_to):
		try:
			for i,x in enumerate(files_from):
				dest=os.path.join(CACHE_DIR,sig[:2],sig,str(i))
				self.bucket_copy(x,dest)
		except Exception:
			return traceback.format_exc()
		return OK
	def copy_from_cache(self,sig,files_from,files_to):
		try:
			for i,x in enumerate(files_to):
				orig=os.path.join(CACHE_DIR,sig[:2],sig,str(i))
				self.bucket_copy(orig,x)
		except EnvironmentError:
			return traceback.format_exc()
		return OK
def loop(service):
	txt=sys.stdin.readline().strip()
	if not txt:
		sys.exit(1)
	ret=OK
	[sig,files_from,files_to]=cPickle.loads(base64.b64decode(txt))
	if files_from:
		ret=service.copy_to_cache(sig,files_from,files_to)
	elif files_to:
		ret=service.copy_from_cache(sig,files_from,files_to)
	else:
		ret="Invalid command"
	obj=base64.b64encode(cPickle.dumps(ret))
	sys.stdout.write(obj.decode())
	sys.stdout.write('\n')
	sys.stdout.flush()
if __name__=='__main__':
	if CACHE_DIR.startswith('s3://')or CACHE_DIR.startswith('gs://')or CACHE_DIR.startswith('minio://'):
		if CACHE_DIR.startswith('minio://'):
			CACHE_DIR=CACHE_DIR[8:]
		service=bucket_cache()
	elif CACHE_DIR.startswith('http'):
		service=netcache()
	else:
		service=fcache()
	while 1:
		try:
			loop(service)
		except KeyboardInterrupt:
			break



# -*- coding: utf-8 -*-

# Copyright the NTPsec project contributors
#
# SPDX-License-Identifier: BSD-2-Clause

from __future__ import print_function

import collections
import inspect
import unittest
import ntp.util
import ntp.packet
import sys
import socket

import jigs


class TestPylibUtilMethods(unittest.TestCase):

    def test_deunicode_units(self):
        u = ntp.util

        self.assertEqual(u.UNIT_US, u"µs")
        self.assertEqual(u.UNIT_PPK, u"‰")
        self.assertEqual(u.UNITS_SEC, ["ns",  u"µs", "ms", "s", "ks"])
        self.assertEqual(u.UNITS_PPX, ["ppt", "ppb", "ppm", u"‰"])
        u.deunicode_units()
        self.assertEqual(u.UNIT_US, "us")
        self.assertEqual(u.UNIT_PPK, "ppk")
        self.assertEqual(u.UNITS_SEC, ["ns",  "us", "ms", "s", "ks"])
        self.assertEqual(u.UNITS_PPX, ["ppt", "ppb", "ppm", "ppk"])
        u.UNIT_US = u"µs"
        u.UNIT_PPK = u"‰"
        u.UNITS_SEC[1] = u.UNIT_US
        u.UNITS_PPX[3] = u.UNIT_PPK

    def test_dolog(self):
        f = ntp.util.dolog

        faketimemod = jigs.TimeModuleJig()
        # We need a test jig

        class LogTester:
            def __init__(self):
                self.written = None
                self.flushed = False

            def write(self, text):
                if self.written is None:
                    self.written = ""
                self.written += text

            def flush(self):
                self.flushed = True
        try:
            timetemp = ntp.util.time
            ntp.util.time = faketimemod
            # Test with logging off (fd == None)
            #   uh... if someone can think of a way to do that please tell me
            # Test with logging on, below threshold
            jig = LogTester()
            faketimemod.time_returns = [0, 1]
            f(jig, "blah", 0, 3)
            self.assertEqual((jig.written, jig.flushed), (None, False))
            # Test with logging on, above threshold
            jig.__init__()  # reset
            f(jig, "blah", 4, 3)
            self.assertEqual((jig.written, jig.flushed),
                             ("1970-01-01T00:00:01Z blah\n", True))
        finally:
            ntp.util.time = timetemp

    def test_safeargcast(self):
        f = ntp.util.safeargcast

        errjig = jigs.FileJig()
        try:
            errtemp = sys.stderr
            sys.stderr = errjig
            # Test successful int
            self.assertEqual(f("42", int, "blah %s", "\nDo the needful\n"), 42)
            self.assertEqual(errjig.data, [])
            # Test successful float
            self.assertAlmostEqual(f("5.23", float, "blah %s",
                                     "\nDo the needful\n"),
                                   5.23)
            self.assertEqual(errjig.data, [])
            # Test failure
            try:
                f("23.5", int, "blah %s", "\nDo the needful\n")
                errored = False
            except SystemExit:
                errored = True
            self.assertEqual(errored, True)
            self.assertEqual(errjig.data, ["blah 23.5", "\nDo the needful\n"])
        finally:
            sys.stderr = errtemp

    def test_stdversion(self):
        f = ntp.util.stdversion

        ver = "@NTPSEC_VERSION_EXTENDED@"
        self.assertEqual(f(), "ntpsec-" + ver)

    def test_rfc3339(self):
        f = ntp.util.rfc3339

        self.assertEqual(f(1480999786),
                         '2016-12-06T04:49:46Z')
        self.assertEqual(f(1480999786.5),
                         '2016-12-06T04:49:46.5Z')
        # RFC 3339, 2 digits of seconds.
        # we round, but the spec is silent on rounding
        # Python 2 and 3 round differently
        if sys.version_info[0] < 3:
            self.assertEqual(f(1480999786.025), "2016-12-06T04:49:46.03Z")
        else:
            self.assertEqual(f(1480999786.025), "2016-12-06T04:49:46.025Z")

    def test_deformatNTPTime(self):
        f = ntp.util.deformatNTPTime

        # Test standard
        self.assertEqual(f("0x0001020304050607.08090A0B0C0D0E0F"),
                         "\x00\x01\x02\x03\x04\x05\x06\x07"
                         "\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F")
        # Test empty
        self.assertEqual(f(""), "")

    def test_hexstr2octets(self):
        f = ntp.util.hexstr2octets

        # Test
        self.assertEqual(f("f00dface"), "\xF0\x0D\xFA\xCE")
        # Test odd length
        self.assertEqual(f("cafebabe1"), "\xCA\xFE\xBA\xBE")

    def test_slicedata(self):
        f = ntp.util.slicedata

        # Test normal
        self.assertEqual(f("foobaz", 2), ("fo", "obaz"))
        # Test overflow
        self.assertEqual(f("foobaz", 8), ("foobaz", ""))
        # Test underflow
        self.assertEqual(f("foobaz", -2), ("foob", "az"))

    def test_portsplit(self):
        self.assertEqual(ntp.util.portsplit("host.invalid"),
                         ("host.invalid", ""))
        self.assertEqual(ntp.util.portsplit("host.invalid:123"),
                         ("host.invalid", ":123"))
        self.assertEqual(ntp.util.portsplit(
            "[0000:1111:2222:3333:4444:5555:6666:7777]:123"),
            ("0000:1111:2222:3333:4444:5555:6666:7777", ":123"))

    def test_parseConf(self):
        data = """foo bar 42 #blah blah blah
        "abcd"   poi? 3.14
        "\\"\\n\\\\foo\\c" \\
        poi! """
        parsed = ntp.util.parseConf(data)
        self.assertEqual(parsed,
                         [[(False, "foo"), (False, "bar"), (False, 42)],
                          [(True, "abcd"), (False, "poi?"), (False, 3.14)],
                          [(True, "\"\n\\foo"), (False, "poi!")]])

    def test_stringfilt(self):
        f = ntp.util.stringfilt

        self.assertEqual(f("1.2345 2.3456 3.4567 4.5678 5.6789 6.789"),
                         " 1.2345  2.3456  3.4567  4.5678  5.6789   6.789")

    def test_oomsbetweenunits(self):
        f = ntp.util.oomsbetweenunits

        self.assertEqual(f(ntp.util.UNIT_KS, ntp.util.UNIT_MS), 6)
        self.assertEqual(f(ntp.util.UNIT_PPM, ntp.util.UNIT_PPB), 3)
        self.assertEqual(f(ntp.util.UNIT_S, None), 9)

    def test_scalestring(self):
        f = ntp.util.scalestring

        # Scale all decimals
        self.assertEqual(f("0.042"), ("42", -1))
        # Unscalable
        self.assertEqual(f(".23"), ("0.23", 0))
        # Typical length, positive value, no scaling
        self.assertEqual(f("1.23450"), ("1.23450", 0))
        # Ditto, negative
        self.assertEqual(f("-1.23450"), ("-1.23450", 0))
        # Long, positive value, no scaling
        self.assertEqual(f("1.234567890123456"), ("1.234567890123456", 0))
        # ditto, negative
        self.assertEqual(f("-1.234567890123456"), ("-1.234567890123456", 0))
        # Zero
        self.assertEqual(f("0.000000"), ("0", -2))
        # Zero with point
        self.assertEqual(f("0.0000"), ("0.0", -1))
        # Negative zero
        self.assertEqual(f("-0.000"), ("0", -1))
        # Large, positive, non scaled
        self.assertEqual(f("987.654"), ("987.654", 0))
        # ditto, negative
        self.assertEqual(f("-987.654"), ("-987.654", 0))
        # Scaled to larger unit, positive
        self.assertEqual(f("12345.67890"), ("12.34567890", 1))
        # ditto, negative
        self.assertEqual(f("-12345.67890"), ("-12.34567890", 1))
        # Scaled to smaller unit, position
        self.assertEqual(f("0.1234567890"), ("123.4567890", -1))
        # ditto, negative
        self.assertEqual(f("-0.1234567890"), ("-123.4567890", -1))
        # Bizarre 1
        self.assertEqual(f("-000.000012345678900987654321"),
                         ("-12.345678900987654321", -2))
        # Bizarre 2
        self.assertEqual(f("1234567890987654321000.00000000000042"),
                         ("1.23456789098765432100000000000000042", 7))
        # Bizarre 3
        self.assertEqual(f("00000000.000000000000"),
                         ("0", -4))

    def test_rescalestring(self):
        f = ntp.util.rescalestring

        # No-op
        self.assertEqual(f("1000.42", 0), "1000.42")
        # Scale to higher unit
        self.assertEqual(f("123.456", 1), "0.123456")
        # ditto, negative
        self.assertEqual(f("-123.456", 1), "-0.123456")
        # Scale to higher unit, beyond available digits
        self.assertEqual(f("12.34", 2), "0.00001234")
        # ditto, negative
        self.assertEqual(f("-12.34", 2), "-0.00001234")
        # Scale to lower unit
        self.assertEqual(f("1.23456", -1), "1234.56")
        # ditto, negative
        self.assertEqual(f("-1.23456", -1), "-1234.56")
        # Scale to lower unit, beyond available digits
        self.assertEqual(f("1.23456", -2), None)
        # ditto, negative
        self.assertEqual(f("-1.23456", -2), None)
        # Scale from below the decimal
        self.assertEqual(f("0.420", -1), "420")

    def test_breaknumberstring(self):
        f = ntp.util.breaknumberstring

        # No decimals, positive
        self.assertEqual(f("1234567890"), ("1234567890", "", False))
        # ditto, negative
        self.assertEqual(f("-1234567890"), ("1234567890", "", True))
        # No whole, positive
        self.assertEqual(f(".12345"), ("", "12345", False))
        # ditto, negative
        self.assertEqual(f("-.12345"), ("", "12345", True))
        # Both sides, position
        self.assertEqual(f("123.456"), ("123", "456", False))
        # ditto, negative
        self.assertEqual(f("-123.456"), ("123", "456", True))

    def test_gluenumberstring(self):
        f = ntp.util.gluenumberstring

        # No decimals, positive
        self.assertEqual(f("123", "", False), "123")
        # ditto, negative
        self.assertEqual(f("123", "", True), "-123")
        # No whole, positive
        self.assertEqual(f("", "456", False), "0.456")
        # ditto, negative
        self.assertEqual(f("", "456", True), "-0.456")
        # Both sides, positive
        self.assertEqual(f("123", "456", False), "123.456")
        # ditto, negative
        self.assertEqual(f("123", "456", True), "-123.456")

    def test_fitinfield(self):
        f = ntp.util.fitinfield

        # Field too small, crop decimals
        self.assertEqual(f("123.456", 5), "123.5")
        # ditto, negative
        self.assertEqual(f("-123.456", 5), "-123.")
        # Field too small, blow field
        self.assertEqual(f("12345.678", 4), "12346.")
        # Goldilocks
        self.assertEqual(f("123.456", 7), "123.456")
        # Field too big
        self.assertEqual(f("123.456", 10), "   123.456")
        # Rounding test, round down
        self.assertEqual(f("1.994", 4), "1.99")
        # Rounding test, round up
        self.assertEqual(f("1.995", 4), "2.00")
        # Attempt to catch bug
        self.assertEqual(f("15937.5", None), "15937.5")
        # Test to bid and no decimals
        self.assertEqual(f("123456", 4), "123456")

    def test_cropprecision(self):
        f = ntp.util.cropprecision

        # No decimals
        self.assertEqual(f("1234", 6), "1234")
        # Decimals, no crop
        self.assertEqual(f("12.3456", 6), "12.3456")
        # Decimals, crop
        self.assertEqual(f("12.3456", 3), "12.345")
        # At baseunit
        self.assertEqual(f("1.234", 0), "1")

    def test_isstringzero(self):
        f = ntp.util.isstringzero

        # Non-zero
        self.assertEqual(f("0.0000001"), False)
        # Zero
        self.assertEqual(f("-0.00"), True)

    def test_unitify(self):
        f = ntp.util.unitify
        nu = ntp.util

        # Zero
        self.assertEqual(f("0.000", nu.UNIT_MS), u"     0µs")
        # Zero, spaced
        self.assertEqual(f("0.000", nu.UNIT_MS, unitSpace=True), u"    0 µs")
        # Standard, width=8
        self.assertEqual(f("1.234", nu.UNIT_MS), " 1.234ms")
        # ditto, negative
        self.assertEqual(f("-1.234", nu.UNIT_MS), "-1.234ms")
        # Scale to larger unit, width=8
        self.assertEqual(f("1234.5", nu.UNIT_MS), " 1.2345s")
        # ditto, negative
        self.assertEqual(f("-1234.5", nu.UNIT_MS), "-1.2345s")
        # Scale to smaller unit, width=8
        self.assertEqual(f("0.01234", nu.UNIT_MS), u" 12.34µs")
        # ditto, negative
        self.assertEqual(f("-0.01234", nu.UNIT_MS), u"-12.34µs")
        # At baseunit
        self.assertEqual(f("12.0", nu.UNIT_NS), "    12ns")
        # Scale to baseunit
        self.assertEqual(f(".042", nu.UNIT_US), "    42ns")
        # Below baseunit
        self.assertEqual(f("23.42", nu.UNIT_NS), "    23ns")
        # Different units
        self.assertEqual(f("12.345", nu.UNIT_PPM), "12.35ppm")
        # Strip
        self.assertEqual(f("1.23", nu.UNIT_MS, width=None), "1.23ms")
        # Different width
        self.assertEqual(f("1.234", nu.UNIT_MS, width=12), "     1.234ms")
        # Outside of available units
        self.assertEqual(f("1234.5", nu.UNIT_KS), "1234.5ks")
        # Seconds
        self.assertEqual(f("42.23", nu.UNIT_S), "  42.23s")
        # Attempt to catch bug
        self.assertEqual(f("15937.5", nu.UNIT_MS, width=None), "15.9375s")

    def test_stringfiltcooker(self):
        # No scale
        self.assertEqual(ntp.util.stringfiltcooker(
            "1.02 34.5 0.67835 -23.0 9 6.7 1.00 .1"),
            "   1.02    34.5 0.67835   -23.0       9     6.7    1.00     "
            "0.1 ms"
        )
        # Scale to larger unit
        self.assertEqual(ntp.util.stringfiltcooker(
            "1000.02 3400.5 0.67835 -23.0 9001 6.7 1.00 1234"),
            "1.00002  3.4005 0.00068 -0.0230   9.001  0.0067 0.00100   "
            "1.234 s"
        )
        # Scale to smaller unit
        self.assertEqual(ntp.util.stringfiltcooker(
            "0.470 0.420 0.430 0.500 0.460 0.4200 0.490 0.480"),
            u"    470     420     430     500     460   420.0     490     "
            u"480 µs")
        # Can't scale
        self.assertEqual(ntp.util.stringfiltcooker(
            "0.47 0.42 0.43 0.50 0.46 0.42 0.49 0.48"),
            "   0.47    0.42    0.43    0.50    0.46    0.42    0.49    "
            "0.48 ms")
        # Can't scale, only one value blocking
        self.assertEqual(ntp.util.stringfiltcooker(
            "0.47 0.4200 0.4300 0.5000 0.4600 0.4200 0.4900 0.4800"),
            "   0.47  0.4200  0.4300  0.5000  0.4600  0.4200  0.4900  "
            "0.4800 ms")

    def test_unitrelativeto(self):
        f = ntp.util.unitrelativeto

        # Scale to smaller unit
        self.assertEqual(f(ntp.util.UNIT_S, -1), ntp.util.UNIT_MS)
        # Scale to larger unit
        self.assertEqual(f(ntp.util.UNIT_S, 1), ntp.util.UNIT_KS)
        # Scale smaller, outside of range
        self.assertEqual(f(ntp.util.UNIT_S, -10), None)
        # Scale larger, outside of range
        self.assertEqual(f(ntp.util.UNIT_NS, 10), None)
        # Get base unit
        self.assertEqual(f(ntp.util.UNIT_KS, None), ntp.util.UNIT_NS)
        # Different unitgroup
        self.assertEqual(f(ntp.util.UNIT_PPM, -1), ntp.util.UNIT_PPB)

    def test_formatzero(self):
        f = ntp.util.formatzero

        # No scaling possible
        self.assertEqual(f("0.00"), ("0.00", 0))
        # Scaling possible
        self.assertEqual(f("0.0000000"), ("0.0", -2))
        # Scaling without a remainder
        self.assertEqual(f("0.000"), ("0", -1))

    def test_unitifyvar(self):
        f = ntp.util.unitifyvar

        # Second var
        self.assertEqual(f("1.234", "tai"), "  1.234s")
        # Millisecond var
        self.assertEqual(f("1.234", "offset"), " 1.234ms")
        # PPM var
        self.assertEqual(f("1.234", "frequency"), "1.234ppm")
        # No unit type
        self.assertEqual(f("1.234", "blahblahblah"), "1.234")

    def test_f8dot4(self):
        f = ntp.util.f8dot4

        # Test signal
        self.assertEqual(f("foo"), "     foo")
        # Test bad type
        self.assertEqual(f({"food": "problematic"}), "       X")
        # Test not a number
        self.assertEqual(f(float("nan")), "     nan")
        # Test Positives
        # Test xxxxxxxx
        self.assertEqual(f(12345678.9), "12345678")
        # Test xxxxxx.x
        self.assertEqual(f(123456.789), "123456.8")
        # Test xxxxx.xx
        self.assertEqual(f(12345.6789), "12345.68")
        # Test xxxx.xxx
        self.assertEqual(f(1234.56789), "1234.568")
        # Test xxx.xxxx
        self.assertEqual(f(123.456789), "123.4568")
        # Test Negatives
        # Test -xxxxxxx
        self.assertEqual(f(-1234567.89), "-1234567")
        # Test -xxxxx.x
        self.assertEqual(f(-12345.6789), "-12345.7")
        # Test -xxxx.xx
        self.assertEqual(f(-1234.56789), "-1234.57")
        # Test -xxx.xxx
        self.assertEqual(f(-123.456789), "-123.457")
        # Test -xx.xxxx
        self.assertEqual(f(-12.3456789), "-12.3457")

    def test_f8dot3(self):
        f = ntp.util.f8dot3

        # Test signal
        self.assertEqual(f("foo"), "     foo")
        # Test bad type
        self.assertEqual(f({"food": "problematic"}), "       X")
        # Test not a number
        self.assertEqual(f(float("nan")), "     nan")
        # Test Positives
        # Test xxxxxxxx
        self.assertEqual(f(12345678.9), "12345678")
        # Test xxxxxx.x
        self.assertEqual(f(123456.789), "123456.8")
        # Test xxxxx.xx
        self.assertEqual(f(12345.6789), "12345.68")
        # Test xxxx.xxx
        self.assertEqual(f(1234.56789), "1234.568")
        # Test Negatives
        # Test -xxxxxxx
        self.assertEqual(f(-1234567.89), "-1234567")
        # Test -xxxxx.x
        self.assertEqual(f(-12345.6789), "-12345.7")
        # Test -xxxx.xx
        self.assertEqual(f(-1234.56789), "-1234.57")
        # Test -xxx.xxx
        self.assertEqual(f(-123.456789), "-123.457")

    def test_Cache(self):
        c = ntp.util.Cache

        monodata = []

        def monoclock_jig():
            return monodata.pop(0)

        # Test init
        cls = c()
        self.assertEqual(cls._cache, {})
        try:
            monotemp = ntp.util.monoclock
            ntp.util.monoclock = monoclock_jig
            # Test set
            monodata = [5, 10, 315, 20]
            cls.set("foo", 42)
            cls.set("bar", 23)
            self.assertEqual(cls._cache, {"foo": (42, 5, 300),
                                          "bar": (23, 10, 300)})
            self.assertEqual(monodata, [315, 20])
            # Test get, expired
            result = cls.get("foo")
            self.assertEqual(result, None)
            self.assertEqual(monodata, [20])
            self.assertEqual(cls._cache, {"bar": (23, 10, 300)})
            # Test get, valid
            result = cls.get("bar")
            self.assertEqual(result, 23)
            self.assertEqual(monodata, [])
            self.assertEqual(cls._cache, {"bar": (23, 10, 300)})
            # Test set, custom TTL
            monodata = [0, 0, 11, 15]
            cls.set("foo", 42, 10)
            cls.set("bar", 23, 20)
            self.assertEqual(cls._cache, {"foo": (42, 0, 10),
                                          "bar": (23, 0, 20)})
            self.assertEqual(monodata, [11, 15])
            # Test get, expired, custom TTL
            result = cls.get("foo")
            self.assertEqual(result, None)
            self.assertEqual(monodata, [15])
            self.assertEqual(cls._cache, {"bar": (23, 0, 20)})
            # Test get, valid, custom TTL
            result = cls.get("bar")
            self.assertEqual(result, 23)
            self.assertEqual(monodata, [])
            self.assertEqual(cls._cache, {"bar": (23, 0, 20)})
        finally:
            ntp.util.monoclock = monotemp

    def test_canonicalize_dns(self):
        f = ntp.util.canonicalize_dns

        fakesockmod = jigs.SocketModuleJig()
        mycache = ntp.util.Cache()
        mycache.set("foo", "bar")
        try:
            cachetemp = ntp.util.canonicalization_cache
            ntp.util.canonicalization_cache = mycache
            sockettemp = ntp.util.socket
            ntp.util.socket = fakesockmod
            # Test cache hit
            self.assertEqual(f("foo"), "bar")
            self.assertEqual(fakesockmod.gai_calls, [])
            # Test addrinfo fail
            fakesockmod.__init__()
            fakesockmod.gai_error_count = 1
            self.assertEqual(f("nothing"), "DNSFAIL:nothing")
            self.assertEqual(fakesockmod.gai_calls,
                             [("nothing", None, 0, 0, 0,
                               socket.AI_CANONNAME)])
            self.assertEqual(fakesockmod.gni_calls, [])
            # Test nameinfo fail
            fakesockmod.__init__()
            fakesockmod.gni_error_count = 1
            fakesockmod.gni_returns = [("www.Hastur.invalid", 42)]
            fakesockmod.gai_returns = [(("family", "socktype", "proto",
                                         "san.Hastur.invalid",
                                         "42.23.%$.(#"),)]
            self.assertEqual(f("bar:42"), "san.hastur.invalid:42")
            # Test nameinfo fail, no canonname
            fakesockmod.__init__()
            mycache.__init__()
            fakesockmod.gni_error_count = 1
            fakesockmod.gni_returns = [("www.Hastur.invalid", 42)]
            fakesockmod.gai_returns = [(("family", "socktype", "proto",
                                         None, "42.23.%$.(#"),)]
            self.assertEqual(f("bar:42"), "bar:42")
            # Test success
            fakesockmod.__init__()
            mycache.__init__()
            fakesockmod.gni_returns = [("www.Hastur.invalid", 42)]
            fakesockmod.gai_returns = [(("family", "socktype", "proto",
                                         None, "42.23.%$.(#"),)]
            self.assertEqual(f("bar:42"), "www.hastur.invalid:42")
        finally:
            ntp.util.canonicalization_cache = cachetemp
            ntp.util.socket = sockettemp

    def test_termsize(self):
        f = ntp.util.termsize

        fakeosmod = jigs.OSModuleJig()
        fakefcntlmod = jigs.FcntlModuleJig()
        fakeshutilmod = jigs.ShutilModuleJig()
        try:
            ostemp = ntp.util.os
            ntp.util.os = fakeosmod
            # Test default
            fakeosmod.isatty_returns = [False]
            self.assertEqual(f(), (80, 24))
            self.assertEqual(fakeosmod.isatty_calls, [1])
            # termsize takes different code paths for different
            # versions of Python
            if str is not bytes:
                # Python 3 version
                try:
                    shutiltemp = ntp.util.shutil
                    ntp.util.shutil = fakeshutilmod
                    fakeosmod.isatty_returns = [True]
                    fakeshutilmod.gts_returns = [(42, 23)]
                    self.assertEqual(f(), (42, 23))
                finally:
                    ntp.util.shutil = shutiltemp
            else:
                # Python 2.x version
                try:
                    import fcntl
                    fcntltemp = fcntl
                    ntp.util.fcntl = fakefcntlmod
                    fakeosmod.isatty_returns = [True]
                    data = ["\x11\x11\x22\x22\x33\x33\x44\x44"]
                    fakefcntlmod.ioctl_returns = data
                    self.assertEqual(f(), (0x2222, 0x1111))
                    self.assertEqual(fakefcntlmod.ioctl_calls,
                                     [(2, ntp.util.termios.TIOCGWINSZ,
                                       "\0\0\0\0\0\0\0\0", False)])
                finally:
                    ntp.util.fcntl = fcntltemp
        finally:
            ntp.util.os = ostemp

    def test_PeerStatusWord(self):
        c = ntp.util.PeerStatusWord

        # Test blank status
        cls = c(0)
        self.assertEqual(cls.event, 0)
        self.assertEqual(cls.event_count, 0)
        self.assertEqual(cls.conf, "no")
        self.assertEqual(cls.reach, "no")
        self.assertEqual(cls.auth, "none")
        self.assertEqual(cls.condition, "reject")
        self.assertEqual(cls.last_event, "")
        # Test max status
        cls = c(0xFFFF)
        self.assertEqual(cls.event, 15)
        self.assertEqual(cls.event_count, 15)
        self.assertEqual(cls.conf, "yes")
        self.assertEqual(cls.reach, "none")
        self.assertEqual(cls.auth, "ok ")
        self.assertEqual(cls.condition, "pps.peer")
        self.assertEqual(cls.last_event, "")
        # Test __str__ of max status
        self.assertEqual(str(cls),
                         "conf=yes, reach=none, auth=ok , "
                         "cond=pps.peer, event= ec=15")
        # Test third options
        cls = c(0x57FF)
        self.assertEqual(cls.event, 15)
        self.assertEqual(cls.event_count, 15)
        self.assertEqual(cls.conf, "no")
        self.assertEqual(cls.reach, "yes")
        self.assertEqual(cls.auth, "bad")
        self.assertEqual(cls.condition, "pps.peer")
        self.assertEqual(cls.last_event, "")
        # Test all newer than OLDVERSION conditions
        cls = c(0x0000)
        self.assertEqual(cls.condition, "reject")
        cls = c(0x0100)
        self.assertEqual(cls.condition, "falsetick")
        cls = c(0x0200)
        self.assertEqual(cls.condition, "excess")
        cls = c(0x0300)
        self.assertEqual(cls.condition, "outlier")
        cls = c(0x0400)
        self.assertEqual(cls.condition, "candidate")
        cls = c(0x0500)
        self.assertEqual(cls.condition, "backup")
        cls = c(0x0600)
        self.assertEqual(cls.condition, "sys.peer")
        cls = c(0x0700)
        self.assertEqual(cls.condition, "pps.peer")
        # Test all older than OLDVERSION conditions
        cls = c(0xF400, 0)
        self.assertEqual(cls.condition, "insane")
        cls = c(0xF800, 0)
        self.assertEqual(cls.condition, "hi_disp")
        cls = c(0xFC00, 0)
        self.assertEqual(cls.condition, "")
        cls = c(0xF100, 0)
        self.assertEqual(cls.condition, "sel_cand")
        cls = c(0xF200, 0)
        self.assertEqual(cls.condition, "sync_cand")
        cls = c(0xF300, 0)
        self.assertEqual(cls.condition, "sys_peer")
        # Test all last events
        cls = c(0xFFF1)
        self.assertEqual(cls.last_event, "mobilize")
        cls = c(0xFFF2)
        self.assertEqual(cls.last_event, "demobilize")
        cls = c(0xFFF3)
        self.assertEqual(cls.last_event, "unreachable")
        cls = c(0xFFF4)
        self.assertEqual(cls.last_event, "reachable")
        cls = c(0xFFF5)
        self.assertEqual(cls.last_event, "restart")
        cls = c(0xFFF6)
        self.assertEqual(cls.last_event, "no_reply")
        cls = c(0xFFF7)
        self.assertEqual(cls.last_event, "rate_exceeded")
        cls = c(0xFFF8)
        self.assertEqual(cls.last_event, "access_denied")
        cls = c(0xFFF9)
        self.assertEqual(cls.last_event, "leap_armed")
        cls = c(0xFFFA)
        self.assertEqual(cls.last_event, "sys_peer")
        cls = c(0xFFFB)
        self.assertEqual(cls.last_event, "clock_alarm")

    def test_cook(self):
        f = ntp.util.cook
        od = collections.OrderedDict

        termsize = (80, 24)

        def termsize_jig():
            return ntp.util.TermSize(*termsize)

        try:
            termtemp = ntp.util.termsize
            ntp.util.termsize = termsize_jig
            # Test empty
            self.assertEqual(f(od()), "\n")
            # Test prettydates
            data = od((("reftime",
                        ("0x00000000.00000000", "0x00000000.00000000")),
                       ("clock",
                        ("0x10000000.00000000", "0x10000000.00000000")),
                       ("org",
                        ("0x20000000.00000000", "0x20000000.00000000")),
                       ("rec",
                        ("0x30000000.00000000", "0x30000000.00000000")),
                       ("xmt",
                        ("0x40000000.00000000", "0x40000000.00000000"))))
            self.assertEqual(f(data),
                   "reftime=00000000.00000000 2036-02-07T06:28:16.000Z,\n"
                   "clock=10000000.00000000 2044-08-10T03:52:32.000Z,\n"
                   "org=20000000.00000000 2053-02-11T01:16:48.000Z,\n"
                   "rec=30000000.00000000 2061-08-14T22:41:04.000Z,\n"
                   "xmt=40000000.00000000 2070-02-15T20:05:20.000Z\n")
            # Test prettydates, with units
            self.assertEqual(f(data, showunits=True),
                   "reftime=00000000.00000000 2036-02-07T06:28:16.000Z,\n"
                   "clock=10000000.00000000 2044-08-10T03:52:32.000Z,\n"
                   "org=20000000.00000000 2053-02-11T01:16:48.000Z,\n"
                   "rec=30000000.00000000 2061-08-14T22:41:04.000Z,\n"
                   "xmt=40000000.00000000 2070-02-15T20:05:20.000Z\n")
            # Test wide terminal
            termsize = (160, 24)
            self.assertEqual(f(data),
                   "reftime=00000000.00000000 2036-02-07T06:28:16.000Z, "
                   "clock=10000000.00000000 2044-08-10T03:52:32.000Z, "
                   "org=20000000.00000000 2053-02-11T01:16:48.000Z,\n"
                   "rec=30000000.00000000 2061-08-14T22:41:04.000Z, "
                   "xmt=40000000.00000000 2070-02-15T20:05:20.000Z\n")
            # Test narrow terminal
            termsize = (40, 24)
            self.assertEqual(f(data),
                   "\nreftime=00000000.00000000 2036-02-07T06:28:16.000Z,\n"
                   "clock=10000000.00000000 2044-08-10T03:52:32.000Z,\n"
                   "org=20000000.00000000 2053-02-11T01:16:48.000Z,\n"
                   "rec=30000000.00000000 2061-08-14T22:41:04.000Z,\n"
                   "xmt=40000000.00000000 2070-02-15T20:05:20.000Z\n")
            termsize = (80, 24)
            # Test ex-obscure cooking
            data = od((("srcadr", ("1.1.1.1", "1.1.1.1")),
                       ("peeradr", ("2.2.2.2", "2.2.2.2")),
                       ("dstadr", ("3.3.3.3", "3.3.3.3")),
                       ("refid", ("blah", "blah"))))
            self.assertEqual(f(data),
                             "srcadr=1.1.1.1, peeradr=2.2.2.2, "
                             "dstadr=3.3.3.3, refid=blah\n")
            # Test ex-obscure cooking
            self.assertEqual(f(data, showunits=True),
                             "srcadr=1.1.1.1, peeradr=2.2.2.2, "
                             "dstadr=3.3.3.3, refid=blah\n")
            # Test leap
            self.assertEqual(f(od((("leap", (0, "0")),)), showunits=True),
                             "leap=00\n")
            self.assertEqual(f(od((("leap", (1, "1")),)), showunits=True),
                             "leap=01\n")
            self.assertEqual(f(od((("leap", (2, "2")),)), showunits=True),
                             "leap=10\n")
            self.assertEqual(f(od((("leap", (3, "3")),)), showunits=True),
                             "leap=11\n")
            # Test leap
            self.assertEqual(f(od((("leap", (0, "0")),))), "leap=00\n")
            self.assertEqual(f(od((("leap", (1, "1")),))), "leap=01\n")
            self.assertEqual(f(od((("leap", (2, "2")),))), "leap=10\n")
            self.assertEqual(f(od((("leap", (3, "3")),))), "leap=11\n")
            # Test leap, with units
            self.assertEqual(f(od((("leap", (0, "0")),)), showunits=True),
                             "leap=00\n")
            self.assertEqual(f(od((("leap", (1, "1")),)), showunits=True),
                             "leap=01\n")
            self.assertEqual(f(od((("leap", (2, "2")),)), showunits=True),
                             "leap=10\n")
            self.assertEqual(f(od((("leap", (3, "3")),)), showunits=True),
                             "leap=11\n")
            # Test reach
            self.assertEqual(f(od((("reach", (1, "1")),))), "reach=001\n")
            # Test reach, with units
            self.assertEqual(f(od((("reach", (1, "1")),)), showunits=True),
                             "reach=001\n")
            # Test specials
            data = od((("filtdelay", ("1 2 3", "1 2 3")),
                       ("filtoffset", ("2 3 4", "2 3 4")),
                       ("filtdisp", ("3 4 5", "3 4 5")),
                       ("filterror", ("4 5 6", "4 5 6"))))
            self.assertEqual(f(data),
                             " filtdelay =1\t2\t3, filtoffset =2\t3\t4,   "
                             "filtdisp =3\t4\t5,  filterror =4\t5\t6\n")
            # Test specials, with units
            self.assertEqual(f(data, showunits=True),
                             " filtdelay =      1       2       3 ms,\n"
                             "filtoffset =      2       3       4 ms,\n"
                             "  filtdisp =      3       4       5 ms,\n"
                             " filterror =      4       5       6 ms\n")
            # Test flash null
            self.assertEqual(f(od((("flash", (0, "0")),))), "flash=00 ok\n")
            # Test flash all bits
            self.assertEqual(f(od((("flash", (65535, "65535")),))),
                             "\nflash=ffff pkt_dup pkt_bogus pkt_unsync "
                             "pkt_denied pkt_auth pkt_stratum pkt_header "
                             "pkt_autokey pkt_crypto peer_stratum "
                             "peer_dist peer_loop peer_unreach\n")
            # Test MS_VARS
            data = od((("rootdelay", (0, "0")), ("rootdisp", (1, "1")),
                       ("offset", (2, "2")), ("sys_jitter", (3, "3")),
                       ("clk_jitter", (4, "4")), ("leapsmearoffset", (5, "5")),
                       ("authdelay", (6, "6")), ("koffset", (7, "7")),
                       ("kmaxerr", (8, "8")), ("kesterr", (9, "9")),
                       ("kprecis", (10, "10")), ("kppsjitter", (11, "11")),
                       ("clk_wander_threshold", (13, "13")),
                       ("tick", (14, "14")), ("in", (15, "15")),
                       ("out", (16, "16")), ("bias", (17, "17")),
                       ("delay", (18, "18")), ("jitter", (19, "19")),
                       ("dispersion", (20, "20")), ("fudgetime1", (21, "21")),
                       ("fudgetime2", (21, "21"))))
            self.assertEqual(f(data), "rootdelay=0, rootdisp=1, offset=2, "
                             "sys_jitter=3, clk_jitter=4,\nleapsmearoffset=5, "
                             "authdelay=6, koffset=7, kmaxerr=8, kesterr=9, "
                             "kprecis=10,\nkppsjitter=11, "
                             "clk_wander_threshold=13, tick=14, in=15, "
                             "out=16, bias=17,\ndelay=18, jitter=19, "
                             "dispersion=20, fudgetime1=21, fudgetime2=21\n")
            # Test MS_VARS, with units
            self.assertEqual(f(data, showunits=True),
                             "rootdelay=0ms, rootdisp=1ms, offset=2ms, "
                             "sys_jitter=3ms, clk_jitter=4ms,\n"
                             "leapsmearoffset=5ms, authdelay=6ms, "
                             "koffset=7ms, kmaxerr=8ms, kesterr=9ms,\n"
                             "kprecis=10ms, kppsjitter=11ms, "
                             "clk_wander_threshold=13ms, tick=14ms, in=15ms,\n"
                             "out=16ms, bias=17ms, delay=18ms, jitter=19ms, "
                             "dispersion=20ms,\nfudgetime1=21ms, "
                             "fudgetime2=21ms\n")
            # Test S_VARS
            data = od((("tai", (0, "0")), ("poll", (1, "1"))))
            self.assertEqual(f(data), "tai=0, poll=1\n")
            # Test S_VARS, with units
            self.assertEqual(f(data, showunits=True), "tai=0s, poll=1s\n")
            # Test PPM_VARS
            data = od((("frequency", (0, "0")), ("clk_wander", (1, "1"))))
            self.assertEqual(f(data), "frequency=0, clk_wander=1\n")
            # Test PPM_VARS, with units
            self.assertEqual(f(data, showunits=True),
                             "frequency=0ppm, clk_wander=1ppm\n")
            # Test unrecognized variable
            data = od((("yeahyeah", (1, "1")),))
            self.assertEqual(f(data), "yeahyeah=1\n")
        finally:
            ntp.util.termsize = termtemp

    def test_MRUSummary(self):
        c = ntp.util.MRUSummary
        e = ntp.packet.MRUEntry

        fakesockmod = jigs.SocketModuleJig()
        mycache = ntp.util.Cache()

        cdns_jig_calls = []
        cdns_jig_returns = []

        def cdns_jig(ip):  # canonicalize_dns()
            cdns_jig_calls.append(ip)
            return cdns_jig_returns.pop(0)

        # Test init
        cls = c(False, debug=3, logfp=sys.stderr)
        self.assertEqual(cls.debug, 3)
        self.assertEqual(cls.logfp, sys.stderr)
        self.assertEqual(cls.now, None)
        self.assertEqual(cls.showhostnames, False)
        self.assertEqual(cls.wideremote, False)
        try:
            socktemp = ntp.util.socket
            ntp.util.socket = fakesockmod
            cachetemp = ntp.util.canonicalization_cache
            ntp.util.canonicalization_cache = mycache
            cdnstemp = ntp.util.canonicalize_dns
            ntp.util.canonicalize_dns = cdns_jig
            # Test summary, first options
            ent = e()
            ent.first = "0x00000000.00000000"
            ent.last = "0xFFFFFFFF.FFFFFFFF"
            ent.ct = 1
            ent.rs = ntp.magic.RES_KOD
            ent.addr = "1.2.3.4:42"
            ent.mv = 0x17
            fakesockmod.gai_returns = [("fam", "type", "proto",
                                        "foo.bar.com", "1.2.3.4")]
            self.assertEqual(cls.summary(ent),
                             "64730 23296      0  400 K 7 2"
                             "      1        -      -    42 1.2.3.4")
            # Test summary, second options
            mycache._cache = {}
            cls.now = 0x00000000
            cls.showhostnames = True
            ent.first = "0x00000100.00000000"
            ent.last = "0x00000200.00000000"
            ent.ct = 65
            ent.rs = ntp.magic.RES_LIMITED
            fakesockmod.gai_returns = [[("fam", "type", "proto",
                                         "foo.bar.com", ("1.2.3.4", 42))]]
            cdns_jig_returns = ["foo.com"]
            self.assertEqual(cls.summary(ent),
                             "64730 23808    256   20 L 7 2     65"
                             "        -      -    42 foo.com")
            # Test summary, third options
            mycache._cache = {}
            ent.ct = 2
            ent.rs = 0
            fakesockmod.gai_error_count = 1
            cdns_jig_returns = ["foobarbaz" * 5]  # 45 chars, will be cropped
            self.assertEqual(cls.summary(ent),
                             "64730 23808    256    0 . 7 2      2        -      -    42"
                             " 1.2.3.4 (foobarbazfoobarbazfoobarbazfoob")
            # Test summary, wide
            mycache._cache = {}
            cls.wideremote = True
            fakesockmod.gai_error_count = 1
            cdns_jig_returns = ["foobarbaz" * 5]  # 45 chars, will be cropped
            self.assertEqual(cls.summary(ent),
                             "64730 23808    256    0 . 7 2      2"
                             "        -      -    42 1.2.3.4 "
                             "(foobarbazfoobarbazfoobarbazfoobarbazfoobarbaz)")
        finally:
            ntp.util.socket = socktemp
            ntp.util.canonicalization_cache = cachetemp
            ntp.util.canonicalize_dns = cdnstemp

    def test_ReslistSummary(self):
        c = ntp.util.ReslistSummary
        m = ntp.util.ReslistSummary._ReslistSummary__getPrefix

        # Test __getPrefix
        #   Test empty
        self.assertEqual(m(""), "/0")
        #   Test base 16
        self.assertEqual(m("FF:FF:F0:00"), "/20")
        #   Test base 10
        self.assertEqual(m("255.240.0.0"), "/12")
        # Test summary
        # Test with everything
        cls = c()
        data = {"hits": "blah", "addr": "42.23.1.2",
                "mask": "FF:FF:0:0", "flags": "qwerty"}
        self.assertEqual(cls.summary(data),
                         "      blah 42.23.1.2/16\n           qwerty\n")
        # Test with missing data
        data = {"addr": "42.23.1.2", "mask": "FF:FF:0:0"}
        self.assertEqual(cls.summary(data), "")

    def test_IfstatsSummary(self):
        c = ntp.util.IfstatsSummary
        od = collections.OrderedDict

        cls = c()
        # Test with all variables available
        data = od((("addr", "1.2.3.4"), ("bcast", "foo"),
                   ("name", "Namus Maximus"), ("en", True), ("flags", 0xFACE),
                   ("rx", 12), ("tx", 34), ("txerr", 56), ("pc", 78),
                   ("up", 90)))
        self.assertEqual(cls.summary(1, data),
                         "  1 Namus Maximus            . face     12     34"
                         "     56    78       90\n    1.2.3.4\n    foo\n")
        # Test without bcast
        data = od((("addr", "1.2.3.4"), ("name", "Namus Maximus"),
                   ("en", True), ("flags", 0xFACE), ("rx", 12), ("tx", 34),
                   ("txerr", 56), ("pc", 78), ("up", 90)))
        self.assertEqual(cls.summary(1, data),
                         "  1 Namus Maximus            . face     12     34"
                         "     56    78       90\n    1.2.3.4\n")
        # Test with missing data
        self.assertEqual(cls.summary(1, od()), "")


class TestPeerSummary(unittest.TestCase):
    target = ntp.util.PeerSummary

    def test___init__(self):
        cls = self.target("peers", 4, True, False, logfp=sys.stderr)
        self.assertEqual(cls.displaymode, "peers")
        self.assertEqual(cls.pktversion, 4)
        self.assertEqual(cls.showhostnames, True)
        self.assertEqual(cls.showunits, False)
        self.assertEqual(cls.wideremote, False)
        self.assertEqual(cls.debug, 0)
        self.assertEqual(cls.logfp, sys.stderr)
        self.assertEqual(cls.termwidth, None)
        self.assertEqual(cls.horizontal_slack, 0)
        self.assertEqual(cls.namewidth, 15)
        self.assertEqual(cls.refidwidth, 15)
        self.assertEqual(cls._PeerSummary__remote, "     remote    ")
        self.assertEqual(cls._PeerSummary__common,
                         "st t when poll reach   delay   offset   ")
        self.assertEqual(cls._PeerSummary__header, None)
        self.assertEqual(cls.polls, [])

    def test_prettyinterval(self):
        m = self.target.prettyinterval

        # Test invalid
        self.assertEqual(m("Failure"), "-")
        # Test <=2048
        self.assertEqual(m(2048), "2048")
        # Test <=300
        self.assertEqual(m(17971), "300m")
        # Test <=96
        self.assertEqual(m(343831), "96h")
        # Test final
        self.assertEqual(m(350000), "4d")

    def test_high_truncate(self):
        m = self.target.high_truncate

        # Test fit
        self.assertEqual(m("foo", 10), "foo")
        # Test doesn't fit
        self.assertEqual(m("lorem ipsum", 5), "-psum")

    def test_is_clock(self):
        m = self.target.is_clock

        # Test True
        data = {"srchost": "(would)"}
        self.assertEqual(m(data), True)
        # Test False, not a clock
        data = {"srchost": "wouldn't"}
        self.assertEqual(m(data), False)
        # Test False, no host
        data = {"blah": "definitely wouldn't"}
        self.assertEqual(m(data), False)

    def test_header(self):
        # Test peers
        cls = self.target("peers", 4, True, False)
        self.assertEqual(cls.header(),
                         "     remote           refid      st t when "
                         "poll reach   delay   offset   jitter")
        # Test opeers
        cls = self.target("opeers", 4, True, False)
        self.assertEqual(cls.header(),
                         "     remote           local      st t when "
                         "poll reach   delay   offset     disp")
        # Test apeers
        cls = self.target("apeers", 4, True, False)
        self.assertEqual(cls.header(),
                         "     remote       refid   assid  st t when "
                         "poll reach   delay   offset   jitter")

    def test_width(self):
        cls = self.target("peers", 4, True, False)
        self.assertEqual(cls.width(), 79)
        cls.horizontal_slack = 10
        self.assertEqual(cls.width(), 89)

    def test_summary(self):
        cls = self.target("peers", 4, True, False)
        cls.header()

        cdns_jig_calls = []
        cdns_jig_returns = []

        def cdns_jig(ip):  # canonicalize_dns()
            cdns_jig_calls.append(ip)
            return cdns_jig_returns.pop(0)

        data = collections.OrderedDict((("delay", (1.234567, "1.234567")),
                                     ("dstadr", ("1.2.3.4", "1.2.3.4")),
                                     ("dstport", ("blah0", "blah0")),
                                     ("filtdelay", ("blah1", "blah1")),
                                     ("filtdisp", ("blah2", "blah2")),
                                     ("filtoffset", ("blah3", "blah3")),
                                     ("flash", ("blah4", "blah4")),
                                     ("headway", ("blah5", "blah5")),
                                     ("hmode", (6, "6")),
                                     ("hpoll", (12, "12")),
                                     ("jitter", (3.14159, "3.14159")),
                                     ("keyid", ("blah6", "blah6")),
                                     ("leap", ("blah7", "blah7")),
                                     ("offset", (2.71828, "2.71828")),
                                     ("pmode", ("blah8", "blah8")),
                                     ("ppoll", (5, "5")),
                                     ("precision", ("blah9", "blah9")),
                                     ("reach", (500, "500")),
                                     ("refid", ("FAIL", "FAIL")),
                                     ("rec", ("0x00000000.00000000",
                                              "0x00000000.00000000")),
                                     ("reftime", ("0x00001000.00000000",
                                                  "0x00001000.00000000")),
                                     ("rootdelay", ("blah10", "blah10")),
                                     ("rootdisp", (299792.458, "299792.458")),
                                     ("srcadr", ("10.20.30.40",
                                                 "10.20.30.40")),
                                     ("srchost", ("15.25.35.45",
                                                  "15.25.35.45")),
                                     ("scrport", ("blah11", "blah11")),
                                     ("stratum", (8, "8")),
                                     ("ttl", ("blah12", "blah12")),
                                     ("unreach", ("blah13", "blah13")),
                                     ("xmt", ("blah14", "blah14")),
                                     ("randomness!", ("foo!", "bar!"))))
        faketimemod = jigs.TimeModuleJig()
        try:
            timetemp = ntp.util.time
            ntp.util.time = faketimemod
            cdnstemp = ntp.util.canonicalize_dns
            ntp.util.canonicalize_dns = cdns_jig
            # Test, no units, hmode=BCLIENTX, peers
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 b 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, hmode=BROADCAST, not multicast
            data["hmode"] = (5, "5")
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 B 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, hmode=BROADCAST, not multicast
            data["srcadr"] = ("224.2.3.4", "224.2.3.4")
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 M 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, hmode=CLIENT, local refclock
            data["srcadr"] = ("10.20.30.40", "10.20.30.40")
            data["hmode"] = (3, "3")
            data["srchost"] = ("(blah)", "(blah)")
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 l 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, hmode=CLIENT, pool
            data["srchost"] = ("15.25.35.45", "15.25.35.45")
            data["refid"] = ("POOL", "POOL")
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .POOL.           8 p 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, hmode=CLIENT, manycast client
            data["srcadr"] = ("224.2.3.4", "224.2.3.4")
            data["refid"] = ("FAIL", "FAIL")
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 a 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, hmode=CLIENT, unicast
            data["srcadr"] = ("10.20.30.40", "10.20.30.40")
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 u 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, hmode=ACTIVE
            data["hmode"] = (1, "1")
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 s 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, hmode=PASSIVE
            data["hmode"] = (2, "2")
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 S 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, don't show hostnames
            cls.showhostnames = False
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#10.20.30.40     .FAIL.           8 S 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, name crop
            cls.showhostnames = True
            cdns_jig_returns = ["clock_canon_blah_jabber_quantum"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon_bla .FAIL.           8 S 6926"
                             "   32  764   1.2346   2.7183   3.1416\n")
            # Test, no units, no name crop
            cls.wideremote = True
            cdns_jig_returns = ["clock_canon_blah_jabber_quantum"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon_blah_jabber_quantum\n"
                             "                 .FAIL.           8 S"
                             " 6926   32  764   1.2346   2.7183   3.1416\n")
            # Test, with units
            cls.showunits = True
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.           8 S 6926"
                             "   32  764 1.2346ms 2.7183ms 3.1416ms\n")
            # Test, low precision formatting with units
            lowpdata = data.copy()
            lowpdata["delay"] = (1.234, "1.234")
            cls.showunits = True
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, lowpdata, 12345),
                             "#clock_canon     .FAIL.           8 S 6926"
                             "   32  764  1.234ms 2.7183ms 3.1416ms\n")
            # Test, low precision formatting with units
            cls.showunits = False
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, lowpdata, 12345),
                             "#clock_canon     .FAIL.           8 S 6926"
                             "   32  764    1.234    2.718    3.142\n")
            # Test, apeers
            cls.showunits = True
            cls.displaymode = "apeers"
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            self.assertEqual(cls.summary(0x500, data, 12345),
                             "#clock_canon     .FAIL.   12345   8 S 6926"
                             "   32  764 1.2346ms 2.7183ms 3.1416ms\n")
            # Test rstatus, previous version
            cls.showunits = True
            cls.displaymode = "apeers"
            cdns_jig_returns = ["clock_canon"]
            faketimemod.time_returns = [0xA0000000]
            cls.pktversion = 0
            self.assertEqual(cls.summary(0x300, data, 12345),
                             "*clock_canon     .FAIL.   12345   8 S 6926"
                             "   32  764 1.2346ms 2.7183ms 3.1416ms\n")
        finally:
            ntp.util.time = timetemp
            ntp.util.canonicalize_dns = cdnstemp

    def test_intervals(self):
        cls = self.target("peers", 4, True, False)
        cls.polls = [1, 2, 3, 4, 5]
        self.assertEqual(cls.intervals(), [1, 2, 3, 4, 5])
        self.assertEqual(cls.polls, [])


class NtpqRvInfoStats(unittest.TestCase):
    """Test functions by controlling their diet and examining output.

    Iterate through casees a 3 tuple, first element is a sub tuple
    it gets fed to the subject. The second is compared to the output.
    The third is a string to append to the test when things break.
    """

    periodic = 230258.509299404568402
    packets = 9586785

    def test_periodize(self):
        """Test ntp.util.periodize by coqtavoric gavage and scatology."""
        # def periodize(period, clipdigits=0)
        cases = (
            ((self.periodic, 5), (round(self.periodic, 5), "2D 15:57:38"), "normal"),
            ((str(self.periodic), 5), (None, "???"), "period is str"),
            (
                (self.periodic, None),
                (round(self.periodic), "2D 15:57:39"),
                "clipdigits is None",
            ),
            # (('there', 'is'),('no', 'cake'),'broca'), # False test
        )
        shot_test(self, ntp.util.periodize, cases)

    def test_packetize(self):
        """Test ntp.util.packetize by coqtavoric gavage and scatology."""
        # def packetize(packets, period, clipdigits=0, periodized=False)
        cases = (
            ((7.5, self.periodic), ("???", "", ""), "packets is float"),
            ((0, self.periodic), (0, "", ""), "packets == 0"),
            (
                (self.packets, str(self.periodic)),
                (self.packets, "", ""),
                "period is str",
            ),
            (
                (self.packets, self.periodic, 0),
                (self.packets, 42.0, "p/s"),
                "more packets, clipdigits == 0",
            ),
            (
                (self.packets, self.periodic, 5),
                (self.packets, 41.63488, "p/s"),
                "more packets, clipdigits = 5",
            ),
            (
                (self.packets // 200, self.periodic, 0, True),
                (self.packets // 200, "00:00:05", "/p"),
                "more time, clipdigits == 5, stringify",
            ),
            (
                (self.packets // 200, self.periodic, 5, True),
                (self.packets // 200, "00:00:04", "/p"),
                "more time, clipdigits == 5, stringify",
            ),
            (
                (self.packets // 200, self.periodic, 0, False),
                (self.packets // 200, 5.0, "s/p"),
                "more time, clipdigits == 5, float",
            ),
            (
                (self.packets // 200, self.periodic, 5, False),
                (self.packets // 200, 4.80376, "s/p"),
                "more time, clipdigits == 5, float",
            ),
            # ((50, 10, 0, True),('the', 'cake', 'lies'),'broca') # False test
        )
        shot_test(self, ntp.util.packetize, cases)

    def test_uptime(self):
        # def uptime(period):
        """Test ntp.util.uptime by coqtavoric gavage and scatology."""
        cases = [
            # [['cake'], 'you will be baked', 'broca'],
            [[-2.6], "23:59:57", "-float"],
            [[0.0], "00:00:00", "0float"],
            [[2.3], "00:00:02", "+float"],
            [[-2], "23:59:58", "-int"],
            [[0], "00:00:00", "0int"],
            [[2], "00:00:02", "+int"],
        ]
        shot_test(self, ntp.util.uptime, cases)


def shot_test(classy, hook, cases):
    """"Run some tests using inputs and outputs from 2.5d iterable.

    Iterate through the cases iterable, and using a hook function to run
    tests with parameters, result, and a message from the child iterable.
    Parameters and results can themselves be iterable.

    Parameters
        ----------
        classy : unittest.TestCase derivative class
            Where the testing happens and where any backing state resides.
        hook : callable
            The function that does the testing.
        cases : iterable
            Where inputs, comparison outputs, and messages are stored.
    """
    if hasattr(classy, "subTest"):

        def st(h, c):
            for s in c:
                with classy.subTest(s=s):
                    classy.assertEqual(h(*s[0]), s[1], msg=s[2])

    else:

        def st(h, c):
            # Does not handle errors well, as a test tool
            l_f, l_s = [], []
            for s in c:
                i = h(*s[0])
                l_f.append([i])
                l_s.append([s[1]])
            classy.assertEqual(l_f, l_s)

    st(hook, cases)


if __name__ == '__main__':
    unittest.main()
